/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  Copyright 2000 Helix Code, Inc. (www.helixcode.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef __SPRUCE_COMPOSER_H__
#define __SPRUCE_COMPOSER_H__

#include <gnome.h>
#include <gmime/gmime.h>

#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus */

#define TYPE_SPRUCE_COMPOSER	       (spruce_composer_get_type ())
#define SPRUCE_COMPOSER(obj)	       (GTK_CHECK_CAST ((obj), TYPE_SPRUCE_COMPOSER, SpruceComposer))
#define SPRUCE_COMPOSER_CLASS(klass)    (GTK_CHECK_CLASS_CAST ((klass), TYPE_SPRUCE_COMPOSER, SpruceComposerClass))
#define IS_SPRUCE_COMPOSER(obj)	       (GTK_CHECK_TYPE ((obj), TYPE_SPRUCE_COMPOSER))
#define IS_SPRUCE_COMPOSER_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((obj), TYPE_SPRUCE_COMPOSER))


typedef struct _SpruceComposer       SpruceComposer;
typedef struct _SpruceComposerClass  SpruceComposerClass;

struct _SpruceComposer {
	GnomeApp parent;
	
	GPtrArray *extra_hdr_names, *extra_hdr_values;
	
	GtkWidget *from;
	GtkWidget *to;
	GtkWidget *cc;
	GtkWidget *bcc;
	GtkWidget *subject;
	GtkWidget *editor;
	GtkWidget *attachments;
	
	char *sig_file;
	
	gboolean send_html;
};

struct _SpruceComposerClass {
	GnomeAppClass parent_class;
	
	void (* send)     (SpruceComposer *composer);
	void (* postpone) (SpruceComposer *composer);
};


GtkType             spruce_composer_get_type          (void);

void                spruce_composer_construct         (SpruceComposer *composer);
SpruceComposer     *spruce_composer_new               (void);
SpruceComposer     *spruce_composer_new_with_sig_file (const char *sig_file);
SpruceComposer     *spruce_composer_new_with_message  (GMimeMessage *msg);
/*SpruceComposer     *spruce_composer_new_from_url      (const char *url);*/

void                spruce_composer_set_headers       (SpruceComposer *composer,
						       const GList *to,
						       const GList *cc,
						       const GList *bcc,
						       const char *subject);
void                spruce_composer_set_body_text     (SpruceComposer *composer,
						       const char *text);
void                spruce_composer_add_header        (SpruceComposer *composer,
						       const char *name,
						       const char *value);
void                spruce_composer_attach            (SpruceComposer *composer,
						       GMimePart *attachment);

GMimeMessage       *spruce_composer_get_message       (SpruceComposer *composer);

void                spruce_composer_set_sig_file      (SpruceComposer *composer,
						       const char *sig_file);
const char         *spruce_composer_get_sig_file      (SpruceComposer *composer);

void                spruce_composer_set_send_html     (SpruceComposer *composer,
						       gboolean send_html);
gboolean            spruce_composer_get_send_html     (SpruceComposer *composer);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __SPRUCE_COMPOSER_H__ */
