/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include "dialog.h"

#include "pixmaps/dialog-default.xpm"
#include "pixmaps/dialog-error.xpm"
#include "pixmaps/dialog-info.xpm"
#include "pixmaps/dialog-question.xpm"
#include "pixmaps/dialog-warning.xpm"

static void on_dialog_destroy (GtkWidget *widget, gpointer user_data);

static void on_ok_button_clicked (GtkWidget *widget, gpointer user_data);


GtkWidget*
spruce_dialog_new (const char *title, const char *description, guint icon)
{
	GtkWidget *dialog;
	GtkWidget *dialog_vbox1;
	GtkWidget *hbox1;
	GtkWidget *lblDescription;
	GtkWidget *dialog_action_area1;
	GtkWidget *hbox2;
	GtkWidget *ok_button;
	GtkWidget *pixmap;
	GdkPixmap *xpm;
	GdkBitmap *mask;
	
	dialog = gtk_dialog_new ();
	gtk_object_set_data (GTK_OBJECT (dialog), "dialog", dialog);
	gtk_widget_set_usize (dialog, 30, 24);
	gtk_window_set_title (GTK_WINDOW (dialog), title);
	gtk_window_set_default_size (GTK_WINDOW (dialog), 450, 150);
	gtk_window_set_policy (GTK_WINDOW (dialog), TRUE, TRUE, FALSE);
	gtk_widget_realize (dialog);
	
	dialog_vbox1 = GTK_DIALOG (dialog)->vbox;
	gtk_object_set_data (GTK_OBJECT (dialog), "dialog_vbox1", dialog_vbox1);
	gtk_widget_show (dialog_vbox1);
	
	hbox1 = gtk_hbox_new (FALSE, 6);
	gtk_widget_ref (hbox1);
	gtk_object_set_data_full (GTK_OBJECT (dialog), "hbox1", hbox1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox1);
	gtk_box_pack_start (GTK_BOX (dialog_vbox1), hbox1, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (hbox1), 5);

	switch (icon) {
	case SPRUCE_DIALOG_PIXMAP_NONE:
		xpm = NULL;
		break;
	case SPRUCE_DIALOG_PIXMAP_ERROR:
		xpm = gdk_pixmap_create_from_xpm_d (dialog->window, &mask,
						    &dialog->style->white, dialog_error_xpm);
		break;
	case SPRUCE_DIALOG_PIXMAP_INFO:
		xpm = gdk_pixmap_create_from_xpm_d (dialog->window, &mask,
						    &dialog->style->white, dialog_info_xpm);
		break;
	case SPRUCE_DIALOG_PIXMAP_QUESTION:
		xpm = gdk_pixmap_create_from_xpm_d (dialog->window, &mask,
						    &dialog->style->white, dialog_question_xpm);
		break;
	case SPRUCE_DIALOG_PIXMAP_WARNING:
		xpm = gdk_pixmap_create_from_xpm_d (dialog->window, &mask,
						    &dialog->style->white, dialog_warning_xpm);
		break;
	default:
		xpm = gdk_pixmap_create_from_xpm_d (dialog->window, &mask,
						    &dialog->style->white, dialog_default_xpm);
	}
	
	if (xpm) {
		pixmap = gtk_pixmap_new (xpm, mask);
		gtk_widget_ref (pixmap);
		gtk_object_set_data_full (GTK_OBJECT (dialog), "pixmap", pixmap,
					  (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (pixmap);
		gtk_box_pack_start (GTK_BOX (hbox1), pixmap, FALSE, FALSE, 0);
	}
	
	lblDescription = gtk_label_new (description);
	gtk_widget_ref (lblDescription);
	gtk_object_set_data_full (GTK_OBJECT (dialog), "lblDescription", lblDescription,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (lblDescription);
	gtk_box_pack_start (GTK_BOX (hbox1), lblDescription, TRUE, TRUE, 0);
	
	dialog_action_area1 = GTK_DIALOG (dialog)->action_area;
	gtk_object_set_data (GTK_OBJECT (dialog), "dialog_action_area1", dialog_action_area1);
	gtk_widget_show (dialog_action_area1);
	gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);
	
	hbox2 = gtk_hbox_new (TRUE, 0);
	gtk_widget_ref (hbox2);
	gtk_object_set_data_full (GTK_OBJECT (dialog), "hbox2", hbox2,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox2);
	gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbox2, TRUE, TRUE, 0);
	
	ok_button = gtk_button_new_with_label (_("OK"));
	gtk_widget_ref (ok_button);
	gtk_object_set_data_full (GTK_OBJECT (dialog), "ok_button", ok_button,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (ok_button);
	gtk_box_pack_start (GTK_BOX (hbox2), ok_button, TRUE, TRUE, 0);
	gtk_widget_set_usize (ok_button, 24, -2);
	
	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
			    GTK_SIGNAL_FUNC (on_dialog_destroy),
			    NULL);
	gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
			    GTK_SIGNAL_FUNC (on_ok_button_clicked),
			    dialog);
	
	return dialog;
}

GtkWidget *
spruce_dialog_new_from_exception (SpruceException *ex)
{
	char *title;
	
	switch (spruce_exception_get_id (ex)) {
	case SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE:
		title = _("Can't autheenticate!");
	default:
		title = _("Warning!");
	}
	
	return spruce_dialog_new (title, spruce_exception_get_description (ex),
				  SPRUCE_DIALOG_PIXMAP_ERROR);
}

void
on_dialog_destroy (GtkWidget *widget, gpointer user_data)
{
	gtk_widget_destroy (widget);
}

void
on_ok_button_clicked (GtkWidget *widget, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET (user_data));
}
