/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  Authors: George Talusan
 *           Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <errno.h>

#include "index.h"

struct Index
{
	guint32 pos;        /* physical offset in mbox */
	guint32 size;       /* physical size of email */
	guint seen;
	gchar *sender;
	gchar *subject;
	gchar *date;
	gchar *msg;
};

Index_t *index_new ()
{
	return g_malloc0(sizeof(Index_t));
}

void index_free (gpointer key, gpointer user_data)
{
	Index_t *index = (Index_t *) key;
	gboolean only_free_msg = GPOINTER_TO_INT (user_data);

	g_return_if_fail (index);

	if (!only_free_msg) {
		if (index->sender) {
			g_free (index->sender);
			index->sender = NULL;
		}
		if (index->subject) {
			g_free (index->subject);
			index->subject = NULL;
		}
		if (index->date) {
			g_free (index->date);
			index->date = NULL;
		}
	}

	if (index->msg) {
		g_free (index->msg);
		index->msg = NULL;
	}

	if (!only_free_msg)
		g_free (index);
}

guint32 index_get_pos (const Index_t *index)
{
	g_return_val_if_fail (index, 0);

	return index->pos;
}

void index_set_pos (Index_t *index, const guint32 pos)
{
	g_assert (index);
	g_return_if_fail (pos != ULONG_MAX && errno != ERANGE);

	index->pos = pos;
}

guint32 index_get_size (const Index_t *index)
{
	g_assert(index);

	return index->size;
}

void index_set_size (Index_t *index, const guint32 size)
{
	g_assert (index);

	index->size = size;
}

gchar *index_get_sender (const Index_t *index)
{
	g_assert (index);

	return index->sender;
}

void index_set_sender (Index_t *index, gchar *sender)
{
	g_assert (index);

	index->sender = g_strdup(sender);
}

gchar *index_get_subject (const Index_t *index)
{
	g_assert (index);

	return index->subject;
}

void index_set_subject (Index_t *index, gchar *subject)
{
	g_assert (index);

	if (subject == '\0')
		/* If we have no subject, show it as empty in the index */
		index->subject = g_strdup("");
	else
		index->subject = g_strdup(subject);
}

gchar *index_get_date (const Index_t *index)
{
	g_assert(index);

	return index->date;
}

void index_set_date (Index_t *index, gchar *date)
{
	g_assert (index);

	index->date = g_strdup(date);
}

gboolean index_get_seen (const Index_t *index)
{
	g_assert(index);

	return index->seen;
}

void index_set_seen (Index_t *index, const gboolean seen)
{
	g_assert (index);

	index->seen = seen;
}

gchar *index_get_msg (const Index_t *index)
{
	g_assert (index);

	return index->msg;
}

void index_set_msg (Index_t *index, gchar *msg)
{
	g_assert (index);

	index->msg = msg;
}

void index_unset_msg (Index_t *index)
{
	if (index->msg)
		index_free(index, GINT_TO_POINTER(TRUE));
}
