/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef __MAILBOX_H__
#define __MAILBOX_H__

#include <glib.h>
#include <pthread.h>
#include "accounts_t.h"
#include "index.h"

#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus */

typedef enum { MAILBOX_LOCAL, MAILBOX_SPOOL, MAILBOX_IMAP } MAILBOX_TYPE;

typedef struct {
	gchar *name;
	MAILBOX_TYPE type;
	ImapAccount_t *imap;
} Mailbox_t;

void mailbox_init (void);

void mailbox_commit (void);

void mailbox_garb_collect (const Mailbox_t *mailbox);

void mailbox_delete (const Mailbox_t *mailbox);

Mailbox_t *mailbox_create (const gchar *mailbox, const MAILBOX_TYPE type);

gchar *mailbox_get_location (const Mailbox_t *mailbox);

Index_t *mailbox_get_index (const Mailbox_t *mailbox, const guint i);

GList *mailbox_get_index_list (const Mailbox_t *mailbox);

gchar *mailbox_get_msg (const Mailbox_t *mailbox, Index_t *index, gboolean *mime, gint trim_level);

gchar *mailbox_get_msg_mime_part (const Mailbox_t *mailbox, Index_t *index, const gint part, gint *mime_part_len);

guint mailbox_get_size (const Mailbox_t *mailbox);

guint mailbox_get_unread (const Mailbox_t *mailbox);

Index_t *mailbox_add_msg_from_file (const Mailbox_t *mailbox, gint fd);

Index_t *mailbox_add_msg (const Mailbox_t *mailbox, const gchar *msg);

gboolean mailbox_del_msgs (const Mailbox_t *mailbox, GList *unsorted_list);

gboolean mailbox_copy_msgs (const Mailbox_t *mbox_dst, const Mailbox_t *mbox_src, GList *unsorted_list);

gboolean mailbox_move_msgs (const Mailbox_t *mbox_dst, const Mailbox_t *mbox_src, GList *unsorted_list);

void mailbox_set_status (const Mailbox_t *mbox, const gboolean status);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
