/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#ifdef WITH_GNOME
#include <gnome.h>
#endif

#include "main.h"
#include "mesg.h"
#include "gti.h"
#include "dialog.h"
#include "spruce-exception.h"

#define d(x) x

/* lets keep track of open windows... */
extern gboolean AboutOpen;
extern gboolean AccountsOpen;
extern gboolean AddrBookOpen;
extern gboolean FiltersOpen;
extern gboolean OptionsOpen;
extern gboolean PrintOpen;
extern gboolean PassphraseOpen;

/* generic globals */
extern gchar *spruceconf;                       /* Spruce config file */
extern gchar *mailboxdir;			/* Spruce mailbox directory */
extern gchar *default_mailboxes[];		/* default mailbox list */
extern gchar *AddrSelected;                     /* addr selected by frmAddrBook */
extern gint first_time;
extern gint texticons;                          /* use text, icons or both? */

extern gint NumFolders;                         /* number of spruce folders */

/* Identity */
extern gchar *SenderAddr;                       /* user's email addr */
extern gchar *ReplyTo;                          /* user's Reply-To email addr */
extern gint use_rand_sig;                       /* use spruce's random signature? */
extern gint auto_attach_sig;                    /* auto-attach the signature? */
extern gchar *signaturefile;                    /* user's signature file */
/* Accounts */
extern GList *PopAccounts;
extern GList *ImapAccounts;
extern Server SmtpServer;                       /* smtp server */
extern gint external_fetchmail;                  /* use an external fetchmail program */
/* Appearance */
extern gchar *mesg_body_font;
extern gchar *unread_mesg_header_font;
extern gchar *read_mesg_header_font;
/* Compose window dimensions, in pixels */
extern gint compose_window_width;
extern gint compose_window_height;
/* Word Wrap */
extern gint word_wrap_len;
extern gint tab_equivalent;
extern gint convert_tabs;
extern gint detect_pgp_clearsigns;
/* Misc */
extern gint save_outgoing;                      /* save outgoing messages? */
extern gint always_prepend;                     /* always prepend Re: or Fw: ?? */
extern gint empty_on_exit;                      /* empty "Deleted items" on exit? */
extern gint save_sizes;                         /* save window and column sizes on exit? */
extern gint cm_delay;                           /* check mail every XX minutes */
extern gint beep_on_new_mail;                   /* beep, when new mails have arrived */
extern gint hdr_trim_level;                     /* 0 = none; 1 = partial; 2 = full */
/* PGP */
extern gchar *pgp_identity;                     /* PGP Identity */
extern gint pgp_sign_always;                    /* pgpv -s ? */
extern gint encrypt_to_self;                    /* always encrypt to self? */
extern gint store_passphrase;                   /* store our passphrase? */
extern gchar *pgp_passphrase;                   /* pgp passphrase */

extern GList *mime_parts;

/* here are our window widgets... */
extern GtkWidget *frmAbout;
extern GtkWidget *frmAccounts;
extern GtkWidget *frmAddrBook;
extern GtkWidget *frmFilters;
extern GtkWidget *frmMain;
extern GtkWidget *frmOptions;
extern GtkWidget *frmPrint;
extern GtkWidget *frmStatus;
extern GtkWidget *dlgPassphrase;

extern GtkWidget *mnuMessage_menu;
extern GtkWidget *mnuMailboxes_menu;
extern GtkWidget *mnuMime_menu;
extern gint sort_column;        /* column to sort */
extern GtkSortType sort_type;

/* used by main.c and gui.c */
static gint cMesg = -1;
static gint NumMesgs = 0;
static gint cMbox = -1;
static GtkWidget *selected_mailbox;
static gboolean refreshing_boxes    = FALSE;
static gboolean refreshing_maillist = FALSE;
static gboolean refreshing_message  = FALSE;
static gboolean MbxNewOpen = FALSE;
static GtkWidget *dlgNewMailbox;
static gchar *SaveFile;
static gboolean SaveAsOpen = FALSE;
static GtkWidget *frmMainSaveAs;
static gboolean fetching_mail = FALSE;
static GtkWidget *dlgCopyMove;
static gboolean CopyMoveOpen = FALSE;
static gint cm_operation = 0;
static gint mime;
static gint mime_preferred_part;

int main (int argc, char *argv[])
{
	GtkWidget *frmComposer = NULL;
	GThreadFunctions *vtable = NULL;
	
	mailbox_init ();
	
	init_thread_funcs (&vtable);
	g_thread_init (vtable);
	
	atexit (global_free);  /* make sure necessary cleanup is taken care of on exit () */
	set_security ();       /* take necessary security precausions */
	reset_globals ();      /* initialize global variables */
	create_default_mailboxes ();
	
	gtk_set_locale ();     /* set language locale */
	
#ifdef ENABLE_NLS
#ifdef HAVE_SETLOCALE_H
	setlocale (LC_ALL,"");
#endif
	bindtextdomain ("spruce", PACKAGE_LOCALE_DIR);
	textdomain ("spruce");
#endif
	
#ifdef WITH_GNOME
	gnome_init ("Spruce", VERSION, argc, argv);
#else
	gtk_init (&argc, &argv);
#endif
	
#ifdef WITH_SSL
	sslInit ();
#endif
	gdk_threads_enter ();
	
	if (!rc_read_defaults ()) {
		frmMain = create_frmMain ();  /* we needa create the main window so we can save defaults */
		rc_save_defaults ();
		
		reset_globals ();
		first_time = 1;
		rc_read_defaults ();  /* seems to crash the Options dialog if we don't do this...hmmm */
	}
	
	spruce_init (argc, argv, &frmComposer); /* read command line after spruce.conf so command line has precedence */
	
	frmMain = create_frmMain ();
	read_accounts ();
	
	read_sizes ();      /* we must read the sizes after frmMain has been created */
	
	filters_read ();    /* read the filter rules */
	
	read_column_sort (); /* read how we're supposed to sort the message lists */
	
	gtk_widget_show (frmMain);
	
	if (frmComposer)
		gtk_widget_show (frmComposer);
	
	if (first_time) {
		frmOptions = create_frmOptions ();
		gtk_widget_show (frmOptions);
		OptionsOpen = TRUE;
	}
	
	refresh_mailboxes (TOP);
	
	/* Check for new mail once every XX minutes. */
	if (cm_delay > 0) {
		gtk_timeout_add (cm_delay * 60 * 1000, periodic_mail_check, 0);
		
		/* Check the mail upon startup, since auto-check is on anyway */
		periodic_mail_check (0);
	}
	
	gtk_main ();
	
	/* program was closed... lets clean up */
	gdk_threads_leave ();
	mailbox_commit ();

	if (empty_on_exit) {
		Mailbox_t *mbox;
		
		mbox = g_malloc0 (sizeof (Mailbox_t));
		mbox->name = DELETED_ITEMS;
		mbox->type = MAILBOX_LOCAL;
		/* empty the contents of DELETED_ITEMS */
		mailbox_delete (mbox);
		g_free (mbox);
		
		mbox = mailbox_create (DELETED_ITEMS, MAILBOX_LOCAL);
		if (mbox) {
			g_free (mbox->name);
			g_free (mbox);
		}
	}
	
	return 0;
}

void on_frmMain_destroy (GtkObject *object, gpointer user_data)
{
	rc_save_defaults (); /* so the window size, and column sizes is saved */
	global_free ();      /* free our global variables */
	gtk_main_quit ();    /* exit */
}

void on_mnuNew_Message_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdNew_clicked (NULL, NULL);
}

void on_cmdNew_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *frmComposer;

	frmComposer = create_frmComposer (SenderAddr, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	gtk_widget_show (frmComposer);
}

void on_mnuFileSave_As_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdSave_clicked (NULL, NULL);
}

void on_cmdSave_clicked (GtkWidget *widget, gpointer data)
{
	g_return_if_fail (!SaveAsOpen);
	g_return_if_fail (cMesg >= 0);

	frmMainSaveAs = create_frmMainSaveAs ();
	gtk_widget_show (frmMainSaveAs);
	SaveAsOpen = TRUE;
}

void on_mnuFilePrint_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdPrintToolbar_clicked (NULL, NULL);
}

void on_cmdPrintToolbar_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	Mailbox_t *mbox;
	Index_t *index;
	gchar *msg, *header;
	gchar *sender, *subject, *date;
	gchar *id;

	g_return_if_fail (!PrintOpen);
	g_return_if_fail (cMesg >= 0);
	g_return_if_fail (selected_mailbox);

	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);

	index = mailbox_get_index(mbox, atoi (id));
	msg = mailbox_get_msg (mbox, index, NULL, hdr_trim_level);

	sender = index_get_sender (index);
	subject = index_get_subject (index);
	date = index_get_date (index);
	
	header = g_strconcat ("From: ", sender ? sender : "", "\n",
			      "Date: ", date ? date : "", "\n",
			      "Subject: ", subject ? subject : "", "\n",
			      NULL);
	frmPrint = create_frmPrint (header, msg);
	g_free (msg);
	g_free (header);
	gtk_widget_show (frmPrint);
	PrintOpen = TRUE;
}

void on_cmdCheckMail_clicked (GtkWidget *widget, gpointer data)
{
	mail_check (SHOW_STATUS_DIALOG);
}

void on_mnuCheck_Mail_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdCheckMail_clicked (NULL, NULL);
}

int invoke_mainthread (gpointer data)
{
	struct invoke_in_gui *i;

	i = (struct invoke_in_gui *) data;
	(*i->func)(i->data);
	g_free (i);

	return 0;
}

/* we have to do it this way, two threads cant write to the X server at
 * the same time */
gint mailthread_refresh_mails (gpointer data); /* <-- forward reference for next function */
void invoke_in_gui_thread (GtkFunction func, gpointer data)
{
	struct invoke_in_gui *i;

	i = g_malloc0 (sizeof (struct invoke_in_gui));
	i->func = func;
	i->data = data;

	/* Execute mail operations in the main thread but do GUI in this one */
	/* (BTW: The mailbox handling routines probably should be made multi-thread
	   safe to avoid any possible problems, especially when background
	   sending is implemented) */
	gdk_threads_enter ();
	if (func == (GtkFunction) mailthread_refresh_mails)
		gtk_idle_add_priority (GTK_PRIORITY_DEFAULT, invoke_mainthread, i);
	else
		invoke_mainthread (i);
	gdk_threads_leave ();
}

gint periodic_mail_check (gpointer data)
{
	mail_check (HIDE_STATUS_DIALOG);
	return 1;
}

void mail_check (gboolean show_dialog)
{
	gint parent;
	gint status;
	pthread_t thread;
	pthread_attr_t thread_attrs;

	if (external_fetchmail) {
		parent = fork ();
		if (!parent) {
			if (execlp ("fetchmail", "fetchmail", NULL) < 0)
				perror ("execl");
		}
		/* FIXME -- Actually do something with this status! */
		wait (&status);
		return;
	}

	if (!fetching_mail) {
		fetching_mail = TRUE;
		
		pthread_attr_init (&thread_attrs);
		pthread_attr_setdetachstate (&thread_attrs, 1);
		
		pthread_create (&thread, &thread_attrs, get_mails, GINT_TO_POINTER (show_dialog));
	}
}

gint mailthread_create_status (gpointer data)
{
	frmStatus = create_frmStatus (NULL);
	gtk_widget_show (frmStatus);

	return 0;
}

gint mailthread_warning_dialog (gpointer data)
{
	GtkWidget *dialog;
	SpruceException *ex = data;
	
	dialog = spruce_dialog_new_from_exception (ex);
	gtk_widget_show (dialog);

	return 0;
}

gint mailthread_destroy_status (gpointer data)
{
	gtk_widget_destroy (frmStatus);

	return 0;
}

gint mailthread_set_status_title (gpointer data)
{
	gtk_window_set_title (GTK_WINDOW (frmStatus), (gchar *) data);

	return 0;
}

gint mailthread_update_status_progress (gpointer data)
{
	GtkWidget *progressbar = get_widget (frmStatus, "progressbar");

	/* Sanity check to prevent crash */
	if (*(gfloat *)data < 0)
		*(gfloat *)data = 0;
	if (*(gfloat *)data > 1)
		*(gfloat *)data = 1;

	gtk_progress_bar_update (GTK_PROGRESS_BAR (progressbar), (*(gfloat *)data));

	return 0;
}

gint mailthread_update_status_text (gpointer data)
{
	GtkWidget *txtStatus;

	txtStatus = get_widget (frmStatus, "txtStatus");

	gtk_entry_set_text (GTK_ENTRY(txtStatus), (gchar *) data);

	return 0;
}

gint mailthread_refresh_mails (gpointer data)
{
	refresh_mail_list ((gint)data);
	
	return 0;
}

gint mailthread_refresh_mailboxes (gpointer data)
{
	refresh_mailboxes_fields ();

	return 0;
}

void *get_mails (gpointer show_dialog)
{
	gboolean show = GPOINTER_TO_INT (show_dialog);
	SpruceException *ex;
	gint i, fetched, total;
	gchar buffer[256];
	gfloat tmp;
	PopAccount_t *pop;
	Mailbox_t mailbox;
	
	if (show) {
		/*frmStatus = create_frmStatus (NULL);
		  gtk_widget_show (frmStatus);*/
		invoke_in_gui_thread (mailthread_create_status, NULL);
	}
	
	total = 0;
	ex = spruce_exception_new ();
	for (i = 0; i < g_list_length (PopAccounts); i++) {
		pop = g_list_nth_data (PopAccounts, i);
		mailbox.type = MAILBOX_LOCAL;
		mailbox.name = pop->mailbox;
		
		g_snprintf (buffer, 255, "%s", pop->server.hostname);
		if (show) {
			/*gtk_window_set_title (GTK_WINDOW (frmStatus), buffer);*/
			invoke_in_gui_thread (mailthread_set_status_title, (gpointer) buffer);
			tmp = 0;
			invoke_in_gui_thread (mailthread_update_status_progress, (gpointer) &tmp);
			/*gtk_progress_bar_update (GTK_PROGRESS_BAR (progressbar), (gfloat) 0);*/
		}
		
		d(fprintf (stderr, "Retrieving mail from: %s:%d...", pop->server.hostname, pop->server.port));
		
		spruce_exception_clear (ex);
		fetched = fetchmail (&pop->server, pop->login, pop->passwd, &mailbox, pop->delemail,
				     show, pop->authtype, pop->useSSL, ex);
		total += fetched;
		
		/*update_display();*/
		
		if (spruce_exception_is_set (ex)) {
			invoke_in_gui_thread (mailthread_warning_dialog, (gpointer) ex);
			/*gtk_widget_show (dialog);*/
			d(fprintf (stderr, "failed.\n%s\n", spruce_exception_get_description (ex)));
		} else {
			d(fprintf (stderr, "done.\n"));
		}
	}
	spruce_exception_free (ex);
	
	if (show)
		invoke_in_gui_thread (mailthread_destroy_status, NULL);
	
	invoke_in_gui_thread (mailthread_refresh_mails, (gpointer) STAY);

	invoke_in_gui_thread (mailthread_refresh_mailboxes, (gpointer) STAY);
	
	if (total > 0 && beep_on_new_mail > 0)
		gdk_beep ();

	fetching_mail = FALSE;
	
	return 0;
}

void on_mnuFileExit_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_frmMain_destroy(NULL, NULL);
}

void on_mnuSettingsAddressBook_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdAddrbk_clicked (NULL, NULL);
}

void on_cmdAddrbk_clicked (GtkWidget *widget, gpointer data)
{
	g_return_if_fail (!AddrBookOpen);

	frmAddrBook = create_frmAddrBook(NULL);
	gtk_widget_show (frmAddrBook);
	AddrBookOpen = TRUE;
}

void on_mnuSettingsAccounts_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!AccountsOpen);

	frmAccounts = create_frmAccounts (TRUE);
	gtk_widget_show (frmAccounts);
	AccountsOpen = TRUE;
}

void on_mnuSettingsFilters_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!FiltersOpen);

	frmFilters = create_frmFilters ();
	gtk_widget_show (frmFilters);
	filters_refresh ();
	FiltersOpen = TRUE;
}

void on_mnuSettingsPreferences_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!OptionsOpen);

	frmOptions = create_frmOptions ();
	gtk_widget_show (frmOptions);
	OptionsOpen = TRUE;
}

void on_mnuMailboxCreate_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!MbxNewOpen);

	dlgNewMailbox = create_dlgNewMailbox();
	gtk_widget_show (dlgNewMailbox);
	MbxNewOpen = TRUE;
}

void on_mnuMailboxDelete_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	GtkWidget *dlgDelMailboxConfirm;

	dlgDelMailboxConfirm = create_dlgDelMailboxConfirm();
	gtk_widget_show (dlgDelMailboxConfirm);
}

void MailboxDelete (void)
{
	Mailbox_t *mbox;
	g_return_if_fail (cMbox >= 0);
	g_return_if_fail (selected_mailbox);

	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);

	mailbox_delete (mbox);

	switch (mbox->type) {
	case MAILBOX_LOCAL:
		NumFolders--;
		if (cMbox >= NumFolders)
			cMbox = NumFolders - 1;
		break;
	case MAILBOX_SPOOL:
		break;
	case MAILBOX_IMAP:
		break;
	default:
		g_assert_not_reached ();
	}

	rc_save_defaults ();

	gtk_widget_destroy(selected_mailbox);
	refresh_mailboxes (TOP);
}

void on_mnuMesgNext_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdNext_clicked (NULL, NULL);
}

void on_cmdNext_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clist = get_widget (frmMain, "clistMail");
	GtkVisibility visible;

	if (cMesg >= NumMesgs - 1)
		return;

	gtk_clist_unselect_row (GTK_CLIST (clist), cMesg, 0);
	gtk_clist_select_row (GTK_CLIST (clist), cMesg + 1, 0);
	visible = gtk_clist_row_is_visible (GTK_CLIST (clist), cMesg);
	if (visible != GTK_VISIBILITY_FULL)
		gtk_clist_moveto (GTK_CLIST (clist), cMesg, 0, 1.0, 0.0);
}

void on_mnuMesgPrevious_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdPrevious_clicked (NULL, NULL);
}

void on_cmdPrevious_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clist = get_widget (frmMain, "clistMail");
	GtkVisibility visible;

	if (cMesg <= 0)
		return;

	gtk_clist_unselect_row (GTK_CLIST (clist), cMesg, 0);
	gtk_clist_select_row (GTK_CLIST (clist), cMesg - 1, 0);
	visible = gtk_clist_row_is_visible (GTK_CLIST (clist), cMesg);
	if (visible != GTK_VISIBILITY_FULL)
		gtk_clist_moveto (GTK_CLIST (clist), cMesg, 0, 0.0, 0.0);
}

void on_mnuMesgReply_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdReply_clicked (NULL, NULL);
}

void on_cmdReply_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clist = get_widget (frmMain, "clistMail");
	GtkWidget *txtBody = get_widget (frmMain, "txtMessage");
	GtkWidget *frmComposer;
	gchar *sender;
	gchar *OrigSubject;
	gchar *subject;
	gchar *body;
	gchar *mod_body, *bodybuffer;
	gchar *replyto;
	gchar *message_id;
	gchar *references, *refs, *refbuffer;
	gchar *id;
	gchar *date;
	gchar *msg;
	Mailbox_t *mbox;
	Index_t *index;
	
	g_return_if_fail (cMesg >= 0);
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 0, &id);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 1, &sender);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 2, &OrigSubject);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 3, &date);
	
	index = mailbox_get_index (mbox, atoi (id));
	g_return_if_fail (index);
	msg = mailbox_get_msg (mbox, index, NULL, 2);
	
	replyto = mesg_get_header_field (msg, "Reply-To:");
	message_id = mesg_get_header_field (msg, "Message-Id:");
	refs = mesg_get_header_field (msg, "References:");
	if (message_id) {
		if (refs) {
			refbuffer = g_strdup_printf ("%s %s", refs, message_id);
			references = word_wrap (refbuffer, 900, 1, FALSE, FALSE);
			g_free (refbuffer);
			g_free (refs);
		} else
			references = g_strdup (message_id);
	} else
		references = refs;
	
	bodybuffer = gtk_editable_get_chars (GTK_EDITABLE (txtBody), 0, -1);
	body = word_wrap (bodybuffer, word_wrap_len - 2, tab_equivalent, convert_tabs, detect_pgp_clearsigns);
	g_free (bodybuffer);
	mod_body = quote_body (body, sender, date);
	
	if (!OrigSubject)
		subject = g_strdup ("Re: ");
	else if (!strstrcase (OrigSubject, "Re: ") || always_prepend)
		subject = g_strconcat ("Re: ", OrigSubject, NULL);
	else
		subject = g_strdup (OrigSubject);
	
	if (!replyto || strstrcase (sender, replyto))
		frmComposer = create_frmComposer (SenderAddr, sender, NULL, NULL, subject,
						  message_id, references, mod_body);
	else
		frmComposer = create_frmComposer (SenderAddr, replyto, NULL, NULL, subject,
						  message_id, references, mod_body);
	gtk_widget_show (frmComposer);
	
	g_free (subject);
	g_free (replyto);
	g_free (message_id);
	g_free (references);
	g_free (mod_body);
	g_free (body);
	g_free (msg);
}

void on_mnuMesgReplyToAll_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdReplyToAll_clicked (NULL, NULL);
}

void on_cmdReplyToAll_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clist   = get_widget (frmMain, "clistMail");
	GtkWidget *txtBody = get_widget (frmMain, "txtMessage");
	GtkWidget *frmComposer;
	gchar *sender      = NULL;
	gchar *OrigSubject = NULL;
	gchar *subject     = NULL;
	gchar *body        = NULL;
	gchar *mod_body    = NULL;
	gchar *bodybuffer  = NULL;
	gchar *replyto     = NULL;
	gchar *message_id  = NULL;
	gchar *references  = NULL;
	gchar *refs        = NULL;
	gchar *refbuffer   = NULL;
	gchar *to          = NULL;
	gchar *tmpto       = NULL;
	gchar *cc          = NULL;
	gchar *date        = NULL;
	gchar *id          = NULL;
	gchar *msg         = NULL;
	Mailbox_t *mbox    = NULL;
	Index_t *index     = NULL;
	
	g_return_if_fail (cMesg >= 0);
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 0, &id);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 1, &sender);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 2, &OrigSubject);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 3, &date);
	
	index = mailbox_get_index (mbox, atoi (id));
	msg = mailbox_get_msg (mbox, index, NULL, 2);
	
	replyto = mesg_get_header_field (msg, "Reply-To:");
	
	message_id = mesg_get_header_field (msg, "Message-Id:");
	refs = mesg_get_header_field (msg, "References:");
	if (message_id) {
		if (refs) {
			refbuffer = g_strdup_printf ("%s %s", refs, message_id);
			references = word_wrap (refbuffer, 900, 1, FALSE, FALSE);
			g_free (refbuffer);
			g_free (refs);
			refs = NULL;
		} else
			references = g_strdup (message_id);
	} else {
		references = refs;
		refs = NULL;
	}
	
	to = mesg_get_header_field (msg, "To:");
	
	if (to != NULL) {
		/* replace '\n' with a single space */
		char *p;
		
		for (p = to; *p; p++)
			if (*p == '\n' || *p == '\t')
				*p = ' ';
		
		while (strrep (to, "  ", " "));
		g_strstrip (to);
		parse_addr_list (to);
	}
	
	cc = mesg_get_header_field (msg, "Cc:");
	
	if (cc != NULL) {
		char *p;
		
		for (p = to; *p; p++)
			if (*p == '\n' || *p == '\t')
				*p = ' ';
		
		while (strrep (cc, "  ", " "));
		g_strstrip (cc);
		parse_addr_list (cc);
	}
	
	bodybuffer = gtk_editable_get_chars (GTK_EDITABLE (txtBody), 0, -1);
	body = word_wrap (bodybuffer, word_wrap_len - 2, tab_equivalent, convert_tabs, detect_pgp_clearsigns);
	g_free (bodybuffer);
	mod_body = quote_body (body, sender, date);
	
	if (!OrigSubject)
		subject = g_strdup ("Re: ");
	else if (!strstrcase (OrigSubject, "Re: ") || always_prepend)
		subject = g_strconcat ("Re: ", OrigSubject, NULL);
	else
		subject = g_strdup (OrigSubject);
	
	if (!replyto || strstrcase (sender, replyto))
		tmpto = to ? g_strconcat (sender, "; ", to, NULL) : g_strdup (sender);
	else
		tmpto = to ? g_strconcat (replyto, "; ", to, NULL) : g_strdup (replyto);

	frmComposer = create_frmComposer (SenderAddr, tmpto, cc, NULL, subject,
					  message_id, references, mod_body);
	gtk_widget_show (frmComposer);
	
	g_free (tmpto);
	g_free (subject);
	g_free (replyto);
	g_free (message_id);
	g_free (references);
	g_free (mod_body);
	g_free (body);
	g_free (to);
	g_free (cc);
	g_free (msg);
}

void on_mnuMesgForward_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdFwd_clicked (NULL, NULL);
}

void on_cmdFwd_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clist   = get_widget (frmMain, "clistMail");
	GtkWidget *txtBody = get_widget (frmMain, "txtMessage");
	GtkWidget *frmComposer;
	gchar *OrigSubject = NULL;
	gchar *subject     = NULL;
	gchar *body        = NULL;
	gchar *mod_body    = NULL;
	gchar *bodybuffer  = NULL;
	gchar *sender      = NULL;
	gchar *date        = NULL;
	
	if (cMesg < 0)
		return;

	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 1, &sender);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 2, &OrigSubject);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 3, &date);

	if (OrigSubject == NULL) {
		/* OrigSubject is NULL */
		subject = g_malloc0(strlen ("Fw: ") + 1);
		sprintf (subject, "Fw: ");
	} else {
		if (!strstrcase (OrigSubject, "Fw: ") || always_prepend) {
			subject = g_malloc0 (strlen ("Fw: ") + 
					     strlen (OrigSubject) + 1);
			sprintf (subject, "Fw: %s", OrigSubject);
		}
		else
			subject = g_strdup (OrigSubject);
	}

	/* wrap by word_wrap_len - 2 because we are gonna add "> " 
	   to the beginning of each line */
	bodybuffer = gtk_editable_get_chars (GTK_EDITABLE (txtBody), 0, -1);
	body = word_wrap (bodybuffer, word_wrap_len - 2, tab_equivalent, convert_tabs, detect_pgp_clearsigns);
	g_free (bodybuffer);
	mod_body = quote_body (body, sender, date);

	frmComposer = create_frmComposer (SenderAddr, NULL, NULL, NULL, 
					  subject, NULL, NULL, mod_body);
	gtk_widget_show (frmComposer);

	g_free (subject);
	g_free (mod_body);
	g_free (body);
}

void on_mnuMesgResend_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	GtkWidget *clist   = get_widget (frmMain, "clistMail");
	GtkWidget *frmComposer;
        GtkWidget *txtBody = get_widget (frmMain, "txtMessage");
	gchar *subject     = NULL;
	gchar *sender      = NULL;
	gchar *to          = NULL;
	gchar *cc          = NULL;
	gchar *replyto     = NULL;
	gchar *message_id  = NULL;
	gchar *references  = NULL;
	gchar *date        = NULL;
	gchar *modbody     = NULL;
	gchar *body        = NULL;
	
	if (cMesg < 0)
		return;
	
	body = gtk_editable_get_chars (GTK_EDITABLE (txtBody), 0, -1);
	
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 1, &sender);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 2, &subject);
	gtk_clist_get_text (GTK_CLIST (clist), cMesg, 3, &date);
	
	to = mesg_get_header_field (body, "To:");
	cc = mesg_get_header_field (body, "Cc:");
	replyto = mesg_get_header_field (body, "Reply-To:");
	message_id = mesg_get_header_field (body, "Message-Id:");
	references = mesg_get_header_field (body, "References:");
	
	modbody = mesg_strip_header (body);
	
	frmComposer = create_frmComposer (sender, to, cc, NULL, subject, 
					  message_id, references, modbody);
	gtk_widget_show (frmComposer);
	
	g_free (to);
	g_free (cc);
	g_free (replyto);
	g_free (message_id);
	g_free (references);
	g_free (modbody);
	g_free (body);
}

void on_mnuMesgOpen_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GtkWidget *frmMesgReader;
	gchar *id, *msg, *subject;
	Mailbox_t *mbox;
	Index_t *index;
	
	g_return_if_fail (selected_mailbox);
	g_return_if_fail (cMesg >= 0);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);
	
	index = mailbox_get_index (mbox, atoi (id));
	msg = mailbox_get_msg (mbox, index, NULL, 2);
	
	subject = index_get_subject (index);
	
	frmMesgReader = create_frmMesgReader (msg, subject ? subject : "");	
	gtk_widget_show (frmMesgReader);
	
	g_free (msg);
}

void on_mnuMesgDelete_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	on_cmdDel_clicked (NULL, NULL);
}

void on_cmdDel_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *clistMail  = get_widget (frmMain, "clistMail");
	gchar *id;
	GList *selection, *del_list = NULL;
	gint i, retval;
	Mailbox_t *mbox, mbox_del;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	selection = GTK_CLIST (clistMail)->selection;
	g_return_if_fail (g_list_length (selection) > 0);
	
	while (selection) {
		gtk_clist_get_text (GTK_CLIST (clistMail), (gint) selection->data, 0, &id);
		del_list = g_list_append (del_list, GINT_TO_POINTER (atoi (id)));
		selection = g_list_next (selection);
	}
	
	/* we should really check the return values here */
	if (strcmp (mbox->name, DELETED_ITEMS)) {
		mbox_del.name = DELETED_ITEMS;
		mbox_del.type = MAILBOX_LOCAL;
		retval = mailbox_move_msgs (&mbox_del, mbox, del_list);
	} else {
		retval = mailbox_del_msgs (mbox, del_list);
	}

	del_list = g_list_first (del_list);
	g_list_free (del_list);
	
	if (retval) {
		clistMail_update ();
		refresh_mailboxes_fields ();
		
		cMesg--;
		
		for (i = 0; i < GTK_CLIST (clistMail)->rows; i++)
			gtk_clist_unselect_row (GTK_CLIST (clistMail), i, 0);
		
		if (NumMesgs > 0)
			gtk_clist_select_row (GTK_CLIST (clistMail), cMesg + 1, 0);
		
		if (!NumMesgs)
		{
			cMesg = -1;
			refresh_txtMessage ();
		}
	}
}

void on_mnuMesgExpunge_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	/* this is more or less a dupe of on_cmdDel_clicked.  clean me up */
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	gchar *id;
	GList *selection, *del_list = NULL;
	guint num = 0;
	gint i;
	Mailbox_t *mbox;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	selection = GTK_CLIST (clistMail)->selection;
	num = g_list_length (selection);
	g_return_if_fail (num > 0);
	
	while (selection) {
		gtk_clist_get_text (GTK_CLIST (clistMail), (gint) selection->data, 0, &id);
		del_list = g_list_append (del_list, GINT_TO_POINTER (atoi (id)));
		selection = g_list_next (selection);
	}

	/* delete messages and update the displayed list assuming it doesn't fail */
	if (mailbox_del_msgs (mbox, del_list)) {
		clistMail_update ();
		refresh_mailboxes_fields ();
		
		cMesg--;
		
		for (i = 0; i < GTK_CLIST (clistMail)->rows; i++)
			gtk_clist_unselect_row (GTK_CLIST (clistMail), i, 0);
		
		if (NumMesgs > 0)
			gtk_clist_select_row (GTK_CLIST (clistMail), cMesg + 1, 0);
		
		if (!NumMesgs) {
			cMesg = -1;
			refresh_txtMessage ();
		}
	}
}

void on_mnuMesgMarkAll_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	gboolean status = GPOINTER_TO_INT (user_data);
	Mailbox_t *mbox;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	mailbox_set_status (mbox, status);
	refresh_mail_list (STAY);
	refresh_mailboxes_fields ();
}

void on_mnuCopy_to_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!CopyMoveOpen);
	g_return_if_fail (cMesg >= 0);
	
	/* show the "Copy To:" dialog */
	cm_operation = MCOPY;
	dlgCopyMove = create_dlgCopyMove ("Copy message to...");
	gtk_widget_show (dlgCopyMove);
	CopyMoveOpen = TRUE;
}

void on_mnuMove_to_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!CopyMoveOpen);
	g_return_if_fail (cMesg >= 0);
	
	/* show the "Move To:" dialog */
	cm_operation = MMOVE;
	dlgCopyMove = create_dlgCopyMove ("Move message to...");
	gtk_widget_show (dlgCopyMove);
	CopyMoveOpen = TRUE;
}

void on_mnuMimeSave_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	GList *tmp;
	GtkWidget *clistMime = get_widget (frmMain, "clistMime");
	GtkWidget *file_sel;
	gchar name[128];
	gint part, ret;
	
	tmp = GTK_CLIST (clistMime)->selection;
	if (tmp == NULL)
		return;
	
	part = (gint)tmp->data;
	tmp = mime_parts;
	while (tmp != NULL && part > 0) {
		tmp=tmp->next;
		part--;
	}
	
	if (((struct mime_part *)tmp->data)->len <= 0)
		return;
	
	file_sel = gtk_file_selection_new (_("Filename"));
	ret = mime_get_parameter_value (((struct mime_part *)tmp->data)->parameter, "name", name);
	if (ret)
		gtk_file_selection_set_filename (GTK_FILE_SELECTION (file_sel), name);
	
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_sel)->ok_button), "clicked", 
			    GTK_SIGNAL_FUNC(on_mnuMimeSave_file), file_sel);
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_sel)->cancel_button), "clicked",
			    GTK_SIGNAL_FUNC(on_mnuMimeSave_file_cancel), file_sel);
	gtk_widget_show (file_sel);
}

void on_mnuMimeSave_file (GtkWidget *button, GtkFileSelection *file_sel)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GtkWidget *clistMime = get_widget (frmMain, "clistMime");
	GList *select;
	FILE *fp;
	Mailbox_t *mbox;
	Index_t *index;
	gchar *id, *mime_part, *mime_file;
	gint mime_part_len;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);
	
	select = GTK_CLIST (clistMime)->selection;
	g_return_if_fail (select);
	
	index = mailbox_get_index(mbox, atoi (id));
	mime_part = mailbox_get_msg_mime_part (mbox, index, (gint)(select->data) + 1,
					       &mime_part_len);
	g_return_if_fail (mime_part);

	mime_file = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_sel));
	
	if ((fp = fopen (mime_file, "wb"))) {
		fwrite (mime_part, 1, mime_part_len, fp);
		fclose (fp);
	} else {
		fprintf (stderr, _("Couldn't open %s\n"), mime_file);
	}
	
	g_free (mime_part);
	gtk_widget_destroy(GTK_WIDGET (file_sel));
}

void on_mnuMimeSave_file_cancel (GtkWidget *button, GtkFileSelection *file_sel)
{
	gtk_widget_destroy(GTK_WIDGET (file_sel));
}

void on_mnuMimeDisplay_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	static gchar tmp_file[80];
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	gchar *id;
	gchar *msg;
	Mailbox_t *mbox;
	Index_t *index;
	FILE *fp;
	gint status, child;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);

	g_return_if_fail (gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id));

	index = mailbox_get_index(mbox, atoi (id));
	msg = mailbox_get_msg (mbox, index, NULL, 0);
	
	strcpy (tmp_file, "/tmp/spruce.XXXXXX");
	if (!mktemp (tmp_file)) {
		perror ("on_mnuMimeDisplay_active::mktemp ()");
		return;
	}

	fp = fopen (tmp_file, "wt");
	chmod (tmp_file, 0600);
	fprintf (fp, "%s\n", msg);
	fclose (fp);
	
	child = fork ();
	switch (child) {
	case -1:
		perror ("Error forking");
		return;
		break;
	case 0:
		execlp ("metamail", "metamail", "-d", tmp_file, NULL);
		perror ("Error executing metamail");
		break;
	default:
		waitpid (child, &status, 0);
	}
	unlink(tmp_file);
}

#ifdef PGP_PROGRAM
void on_mnuPgpPassphrase_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	/* get the passphrase */
	if (!PassphraseOpen) {
		dlgPassphrase = create_dlgPassphrase (NULL);
		gtk_widget_show (dlgPassphrase);
		PassphraseOpen = TRUE;
	}
}

void on_mnuPgpDestroy_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_free (pgp_passphrase);
	pgp_passphrase = NULL;
}
#endif /* PGP_PROGRAM */

void on_mnuHelpAbout_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	g_return_if_fail (!AboutOpen);
	
	/* show the about dialog box... */
	frmAbout = create_frmAbout ();
	gtk_widget_show (frmAbout);
	AboutOpen = TRUE;
}

void on_mnuHelpManual_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	/* bah, not right now */
}

void on_txtMessage_clicked (GtkWidget *text, GdkEvent *event, gpointer user_data)
{
	/* mouse wheel scroll up */
	if (event->button.button == 5) {
		gtk_adjustment_set_value (GTK_ADJUSTMENT (GTK_TEXT (text)->vadj),
					  GTK_ADJUSTMENT (GTK_TEXT (text)->vadj)->value+15); 
	}
	
	/* mouse wheel scroll down */
	if (event->button.button == 4) {
		gtk_adjustment_set_value (GTK_ADJUSTMENT (GTK_TEXT (text)->vadj),
					  GTK_ADJUSTMENT (GTK_TEXT (text)->vadj)->value-15);
	}
}

void on_clistMail_select_row (GtkCList *clist, gint row, gint column, GdkEvent *event, gpointer user_data)
{
	GList *selection;
	
	selection = GTK_CLIST (clist)->selection;
	/* there's no reason to refresh txtMessage if there are
	 * more then one message selected */
	if (g_list_length (selection) == 1 && cMesg != row) {
		/* highlight/select the new row */
		cMesg = row;
		refresh_txtMessage ();
		update_messageStatus ();
	}
}

void on_clistMail_clicked (GtkWidget *clist, GdkEvent *event, gpointer user_data)
{
	if (event->button.button == 1 && (event->type == GDK_2BUTTON_PRESS ||
					  event->type == GDK_3BUTTON_PRESS)) {
		on_mnuMesgOpen_activate (NULL, NULL);
	} else if (event->button.button == 3) {
		gtk_menu_popup ((GtkMenu *)mnuMessage_menu, NULL, NULL, NULL,
				NULL, 0, 0);
	}
}

void on_clistMail_drag_data_get (GtkWidget *clist, GdkDragContext *context, 
                                 GtkSelectionData *selection, guint info, 
                                 guint32 clk_time, gpointer user_data)
{
	gchar string[32];
	
	strcpy (string, "data on the way");
	gtk_selection_data_set (selection, selection->target, 8, string, strlen (string));
}

void on_clistMail_click_column (GtkWidget *clist, gint column, gpointer user_data)
{
	gtk_clist_freeze (GTK_CLIST (clist));
	
	if (column == sort_column) {
		if (sort_type == GTK_SORT_ASCENDING)
			sort_type = GTK_SORT_DESCENDING;
		else
			sort_type = GTK_SORT_ASCENDING;
		gtk_clist_set_sort_type (GTK_CLIST (clist), sort_type);
	} else {
		sort_column = column;
		gtk_clist_set_sort_column (GTK_CLIST (clist), sort_column);
		sort_type = GTK_SORT_ASCENDING;
		gtk_clist_set_sort_type (GTK_CLIST (clist), sort_type);
		switch (sort_column) {
		case 0:         /* sort by message id */
			gtk_clist_set_compare_func (GTK_CLIST (clist), id_compare);
			break;
		case 1:         /* sort by sender */
			gtk_clist_set_compare_func (GTK_CLIST (clist), str_compare);
			break;
		case 2:         /* sort by subject */
			gtk_clist_set_compare_func (GTK_CLIST (clist), subject_compare);
			break;
		case 3:         /* sort by date */
			gtk_clist_set_compare_func (GTK_CLIST (clist), date_compare);
			break;
		default:
			gtk_clist_set_compare_func (GTK_CLIST (clist), str_compare);
			break;
		}
	}
	
	gtk_clist_sort (GTK_CLIST (clist));
	
	/* Reset the message pointer and move the view to the selected one */
	if (GTK_CLIST (clist)->selection) {
		cMesg = (gint) GTK_CLIST (clist)->selection[0].data;
		gtk_clist_moveto (GTK_CLIST (clist), cMesg, -1, -1, -1);
	}
	
	gtk_clist_thaw (GTK_CLIST (clist));
}

void on_treeBoxes_selection_changed (GtkWidget *tree)
{
	GList *selection;
	Mailbox_t *mbox;
	
	selection = GTK_TREE_SELECTION (tree);
	if (selection == NULL) {
		selected_mailbox = NULL;
		return;
	}
	
	if (GTK_TREE_ITEM (GTK_WIDGET (selection->data))->subtree == NULL &&
	    selection->data != selected_mailbox) {
		if (selected_mailbox) {
			mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
			mailbox_garb_collect (mbox);
		}
		
		selected_mailbox = selection->data;
		
		refresh_mail_list (TOP);
	}
}

/*void on_clistBoxes_select_row (GtkCList *clist, gint row, gint column,
                               GdkEvent *event, gpointer user_data)
{
   cMbox = row;
   refresh_mail_list (TOP);
}*/

void on_treeBoxes_clicked (GtkWidget *tree, GdkEvent *event, gpointer user_data)
{
	if (event->button.button == 3) {
		gtk_menu_popup ((GtkMenu *)mnuMailboxes_menu, NULL, NULL, NULL, 
				(gpointer)NULL, (guint) NULL, (guint32)NULL);
	}
}

void on_treeBoxes_drag_data_recieved (GtkWidget *item, GdkDragContext *context, 
				      gint x, gint y, 
				      GtkSelectionData *selection_data, 
				      guint info, guint32 clk_time)
{
#if 0
	gchar *mailbox, *mesgid, *oldmailbox;
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GList *selection;
	gint num = 0, *mesg_nums, i = 0, done, temp;
	struct mailboxformat *mb_old, *mb_new;
	
	/*gtk_clist_get_selection_info (GTK_CLIST (clist), x, y, &row, &col);
	  gtk_clist_get_text (GTK_CLIST (clist), row, 0, &mailbox);
	  gtk_clist_get_text (GTK_CLIST (clist), cMbox, 0, &oldmailbox);*/
	
	fprintf (stderr, "We got a drag&drop signal...\n");
	
	if (selected_mailbox == NULL)
		return;
	
	mb_old = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	if (mb_old == NULL)
		return;
	
	gtk_label_get (GTK_LABEL (GTK_BIN (GTK_WIDGET (selected_mailbox))->child), &oldmailbox);
	
	mb_new = gtk_object_get_data (GTK_OBJECT (item), "mailboxformat");
	if (mb_new == NULL)
		return;
	
	gtk_label_get (GTK_LABEL (GTK_BIN (GTK_WIDGET (item))->child), &mailbox);
	
	
	if (!strcmp (mailbox, oldmailbox)) {
		/* mailboxes are the same */
		fprintf (stderr, "source and destination mailboxes are the same...skipping.\n");
		return;
	}
	
	selection = GTK_CLIST (clistMail)->selection;
	while (selection != NULL) {
		num++;
		selection = selection->next;
	}
	
	mesg_nums = g_malloc0(sizeof (gint)*num);
	
	selection = GTK_CLIST (clistMail)->selection;
	while (selection != NULL) {
		gtk_clist_get_text (GTK_CLIST (clistMail), (gint)selection->data, 0, &mesgid);
		mesg_nums[i] = atoi (mesgid);
		i++;
		selection=selection->next;
	}

	done = 0;
	while (done == 0) {
		done = 1;
		i = 0;
		while (i < num - 1) {
			if (mesg_nums[i] > mesg_nums[i+1]) {
				temp = mesg_nums[i];
				mesg_nums[i] = mesg_nums[i+1];
				mesg_nums[i+1] = temp;
				done = 0;
			}
			i++;
		}
	}

	if (context->action == GDK_ACTION_MOVE) {
		if (mb_new->format == 1 && mb_old->format == 1) {
			mailbox_move_mesgs (num, mesg_nums, oldmailbox, mailbox);
			mailbox_refresh_mailpos (num, mesg_nums, NumMesgs, oldmailbox);
			clistMail_update ();
		}
	}

	if (context->action == GDK_ACTION_COPY) {
		if (mb_new->format == 1 && mb_old->format == 1)
			mailbox_copy_mesgs (num, mesg_nums, oldmailbox, mailbox);
	}
   
	g_free (mesg_nums);
	gtk_drag_finish (context, 1, FALSE, clk_time);
#endif
}

void on_clistMime_select_row (GtkWidget *clist, gint row, gint col, gpointer user_data)
{
	gint i;
	GList *tmp;
	
	i = 0;
	tmp = mime_parts;
	while (tmp != NULL && i < row) {
		tmp=tmp->next;
		i++;
	}
	
	if (!g_strcasecmp (((struct mime_part *)tmp->data)->type, "text") ||
	    !g_strcasecmp (((struct mime_part *)tmp->data)->type, "message")) {
		mime_preferred_part = row + 1;
		refresh_txtMessage ();
	}
}

void on_clistMime_button_press_event (GtkWidget *clist, GdkEvent *event, gpointer user_data)
{
	if (event->button.button == 3) {
		gtk_menu_popup (GTK_MENU (mnuMime_menu), NULL, NULL, NULL,
				(gpointer)NULL, (guint)NULL, (guint32)NULL);
	}
}

void on_frmMainSaveAs_destroy (GtkObject *object, gpointer user_data)
{
	gtk_widget_destroy(frmMainSaveAs);
	SaveAsOpen = 0;
}

void on_cmdMainSaveAs_OK_clicked (GtkButton *button, gpointer user_data)
{
	if (SaveFile != NULL)
		g_free (SaveFile);
	SaveFile = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (frmMainSaveAs)));
	
	save_message ();
	
	gtk_widget_destroy(frmMainSaveAs);
	SaveAsOpen = FALSE;
}

void on_cmdMainSaveAs_Cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy(frmMainSaveAs);
	SaveAsOpen = FALSE;
}

void on_dlgNewMailbox_destroy (GtkObject *object, gpointer user_data)
{
	gtk_widget_destroy(dlgNewMailbox);
	MbxNewOpen = FALSE;
}

/* currently you can't create remote mailboxes */
void on_cmdDlgNewMailboxOK_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *txtMailboxName = get_widget (dlgNewMailbox, "txtMailboxName");
	GtkWidget *radSpruce = get_widget (dlgNewMailbox, "radSpruce");
	GtkWidget *treelocalBoxes = get_widget (frmMain, "treelocalBoxes");
	GtkWidget *treespoolBoxes = get_widget (frmMain, "treespoolBoxes");
	GtkWidget *item = NULL;
	gchar *mailbox;
	Mailbox_t *mbox;

	mailbox = gtk_entry_get_text (GTK_ENTRY(txtMailboxName));
	g_return_if_fail (strlen (mailbox) > 0);

	if (GTK_TOGGLE_BUTTON (radSpruce)->active) {
		if (!GTK_IS_TREE (treelocalBoxes)) {
			fprintf (stderr, "Error: Local Mailbox Tree is NULL.  Cannot append a new mailbox.\n");
			gtk_widget_destroy(dlgNewMailbox);
			MbxNewOpen = FALSE;
			return;
		}

		item = gtk_tree_item_new_with_label (mailbox);
		gtk_tree_append (GTK_TREE (treelocalBoxes), item);
		mbox = mailbox_create (mailbox, MAILBOX_LOCAL);
		gtk_object_set_data (GTK_OBJECT (item), "mailboxformat", mbox);
		gtk_signal_connect (GTK_OBJECT (item), "button_press_event",
				    GTK_SIGNAL_FUNC(on_treeBoxes_clicked), NULL);
		NumFolders++;
	} else {
		if (!GTK_IS_TREE (treespoolBoxes)) {
			fprintf (stderr, "Error: Local Spool Tree is NULL. Cannot append a new mailbox.\n");
			gtk_widget_destroy(dlgNewMailbox);
			MbxNewOpen = FALSE;
			return;
		}
		
		item = gtk_tree_item_new_with_label (mailbox);
		gtk_tree_append (GTK_TREE (treespoolBoxes), item);
		mbox = mailbox_create (mailbox, MAILBOX_SPOOL);
		gtk_object_set_data (GTK_OBJECT (item), "mailboxformat", mbox);
		gtk_signal_connect (GTK_OBJECT (item), "button_press_event",
				    GTK_SIGNAL_FUNC(on_treeBoxes_clicked), NULL);
	}
	
	rc_save_defaults ();
	refresh_mailboxes (STAY);
	
	gtk_widget_destroy(dlgNewMailbox);
	MbxNewOpen = FALSE;
}

void on_cmdDlgNewMailboxCancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy(dlgNewMailbox);
	MbxNewOpen = FALSE;
}

void on_dlgCopyMove_destroy (GtkObject *object, gpointer user_data)
{
	gtk_widget_destroy(dlgCopyMove);
	CopyMoveOpen = FALSE;
}

void on_cmdCopyMoveOK_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *cmboMailbox = get_widget (dlgCopyMove, "cmboMailbox");
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GList *selection, *list = NULL;
	gchar *id;
	Mailbox_t *mbox, mbox_dst;
	
	if (!selected_mailbox) {
		gtk_widget_destroy(dlgCopyMove);
		CopyMoveOpen = FALSE;
		return;
	}

	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");

	if (!mbox) {
		gtk_widget_destroy(dlgCopyMove);
		CopyMoveOpen = FALSE;
		return;
	}

	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);
	
	mbox_dst.name = gtk_entry_get_text (GTK_ENTRY(GTK_COMBO (cmboMailbox)->entry));
	mbox_dst.type = MAILBOX_LOCAL; /* only local for now */

	if (!strcmp (mbox->name, mbox_dst.name)) {
		gtk_widget_destroy(dlgCopyMove);
		CopyMoveOpen = FALSE;
		return;
	}

	selection = GTK_CLIST (clistMail)->selection;

	if (g_list_length (selection) <= 0) {
		gtk_widget_destroy(dlgCopyMove);
		CopyMoveOpen = FALSE;
		return;
	}

	while (selection) {
		gtk_clist_get_text (GTK_CLIST (clistMail), (gint) selection->data, 0, &id);
		list = g_list_append (list, GINT_TO_POINTER (atoi (id)));
		selection = g_list_next (selection);
	}

	switch (cm_operation) {
	case MCOPY:
		mailbox_copy_msgs (&mbox_dst, mbox, list);
		refresh_mailboxes_fields ();
		break;
	case MMOVE:
		mailbox_move_msgs (&mbox_dst, mbox, list);
		refresh_mail_list (STAY);
		refresh_mailboxes_fields ();
		break;
	}
	list = g_list_first (list);
	g_list_free (list);

	gtk_widget_destroy (dlgCopyMove);
	CopyMoveOpen = FALSE;
}

void on_cmdCopyMoveCancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy(dlgCopyMove);
	CopyMoveOpen = FALSE;
}

void on_dlgDelMailbox_destroy (GtkWidget *widget, gpointer user_data)
{
	gtk_widget_destroy(widget);
}

void on_cmdDelMailbox_Yes_clicked (GtkWidget *widget, gpointer user_data)
{
	gtk_widget_destroy(widget);
	MailboxDelete ();
}

void on_cmdDelMailbox_No_clicked (GtkWidget *widget, gpointer user_data)
{
	gtk_widget_destroy(widget);
}

void on_imaptree_expand (GtkWidget *item, Mailbox_t *mbox)
{
	gint sock;
	GList *mailboxes = NULL;
	GList *tmp;
	GtkWidget *mailbox_item;
	ImapAccount_t *imap;
	
	g_return_if_fail (mbox);
	imap = mbox->imap;
	sock = imap_connect (&imap->server, imap->useSSL);

	if (sock < 0) {
		fprintf (stderr, _("Couldn't connect to imap server %s\n"), imap->server.hostname);
		fflush (stderr);
		return;
	}

	imap->socket = sock;
	
	switch (imap->authtype) {
	case PLAIN:
		if (!imap_login (sock, imap->login, imap->passwd)) {
			fprintf (stderr, _("Couldn't log in to imap server %s\n"),
				 imap->server.hostname);
			fflush (stderr);
		}
		break;
	case CRAM_MD5:
		if (!imap_login_cram_md5 (sock, imap->login, imap->passwd)) {
			fprintf (stderr, _("Couldn't log in to imap server %s using CRAM-MD5\n"),
				 imap->server.hostname);
			fflush (stderr);
		}
		break;
	case APOP:
		fprintf (stderr, "APOP is not a valid authentication for the IMAP protocol.\n");
		fflush (stderr);
		break;
	default:
		fprintf (stderr, "Illegal IMAP authentication type.\n");
		return;
	}
	
	mailboxes = imap_list (sock, imap->namespace);
	mailboxes = g_list_insert (mailboxes, "INBOX", 0);
	
	tmp = mailboxes;
	while (tmp != NULL) {
		Mailbox_t *mbox_child = g_memdup (mbox, sizeof (Mailbox_t));
		
		mailbox_item = gtk_tree_item_new_with_label (tmp->data);
		mbox_child->name = tmp->data;
		
		gtk_widget_show (mailbox_item);
		gtk_object_set_data (GTK_OBJECT (mailbox_item), "mailboxformat", mbox_child);
		gtk_tree_append (GTK_TREE (GTK_TREE_ITEM (item)->subtree), mailbox_item);
		tmp = tmp->next;
	}

	g_list_free (mailboxes);
}

void on_imaptree_collapse (GtkWidget *item, Mailbox_t *mbox)
{
	GList *children;
	GtkWidget *tree;
	ImapAccount_t *imap;
	
	g_return_if_fail (mbox);
	imap = mbox->imap;
	g_return_if_fail (imap->socket >= 0);
	imap_logout (imap->socket);
	close (imap->socket);
	
	imap->socket = -1;
	
	children = gtk_container_children (GTK_CONTAINER (GTK_TREE_ITEM (item)->subtree));
	while (children) {
		gtk_object_remove_data (GTK_OBJECT (children->data), "mailboxformat");
		gtk_container_remove (GTK_CONTAINER (GTK_TREE_ITEM (item)->subtree), children->data);
		children = g_list_remove_link (children, children);
	}

	tree = gtk_tree_new ();
	gtk_tree_item_set_subtree (GTK_TREE_ITEM (item), tree);
	
	refresh_mail_list (TOP);
	cMesg = -1;
	refresh_txtMessage ();
}

gint save_message (void)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	gchar *id, *msg;
	Mailbox_t *mbox;
	Index_t *index;
	FILE *fp;
	gint i;
	
	g_return_val_if_fail (selected_mailbox, -1);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_val_if_fail (mbox, -1);
   
	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);
	
	fp = fopen (SaveFile, "wt");
	g_return_val_if_fail (fp, 0);
	
	index = mailbox_get_index(mbox, atoi (id));
	msg = mailbox_get_msg (mbox, index, NULL, hdr_trim_level);

	if (hdr_trim_level == HEADER_TRIMLEVEL_KEEP_NONE) {
		gchar *sender, *subject, *date;
		
		sender =  index_get_sender (index);
		subject =  index_get_subject (index);
		date = index_get_date (index);

		fprintf (fp, "Date:    %s\n", date ? date : "");
		fprintf (fp, "From:    %s\n", sender ? sender : "");
		fprintf (fp, "Subject: %s\n", subject ? subject : "");
	}
	
	for (i = 0; i < 78; i++)
		fprintf (fp, "-");
	fprintf (fp, "\n\n");
	
	fprintf (fp, "%s", msg);
	
	fclose (fp);
	g_free (msg);
	
	return 1;
}

void clear_mailbox_tree (GtkWidget *tree_to_clear, gchar *treename)
{
	GtkWidget *tree;
	GList *children;

	children =
	    gtk_container_children (GTK_CONTAINER
				    (GTK_TREE_ITEM (tree_to_clear)->
				     subtree));
	while (children) {
		gtk_object_remove_data (GTK_OBJECT (children->data),
					"mailboxformat");
		gtk_container_remove (GTK_CONTAINER
				      (GTK_TREE_ITEM (tree_to_clear)->
				       subtree), children->data);
		children = g_list_remove_link (children, children);
	}

	if (GTK_TREE_ITEM (tree_to_clear)->subtree == NULL) {
		tree = gtk_tree_new ();
		gtk_tree_item_set_subtree (GTK_TREE_ITEM (tree_to_clear),
					   tree);
		gtk_object_set_data (GTK_OBJECT (frmMain),
				     treename, tree);
		gtk_tree_item_expand (GTK_TREE_ITEM (tree_to_clear));
	}

}

void add_local_mailbox_to_tree (char *mailbox, gpointer user_data)
{
	GtkWidget *treelocalBoxes =
		get_widget (frmMain, "treelocalBoxes_item");
	add_mailbox_to_tree (treelocalBoxes, MAILBOX_LOCAL, mailbox);
}

void add_mailbox_to_tree (GtkWidget *tree, MAILBOX_TYPE mboxtype, 
			  gchar *mboxname)
{
	GtkWidget *item, *hbox, *label, *num_mails, *num_unread_mails;
	Mailbox_t *mbox;
	char tmp[64];
	int i, unread_mails;
	GtkTargetEntry target[] =
	    { {"STRING", 0, 1}, {"text/plain", 0, 2} };
	
	mbox = g_malloc0 (sizeof (Mailbox_t));
	mbox->type = mboxtype;
	mbox->name = g_strdup (mboxname);
	item = gtk_tree_item_new ();
	
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_set_name (hbox, "hbox");
	gtk_widget_show (hbox);
	gtk_container_add (GTK_CONTAINER (item), hbox);

	label = gtk_label_new (mboxname);
	gtk_widget_set_name (label, "label");
	gtk_object_set_data (GTK_OBJECT (item), "label", label);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);

	sprintf (tmp, "%d", mailbox_get_size (mbox));
	num_mails = gtk_label_new (tmp);
	gtk_widget_set_name (num_mails, "num_mails");
	gtk_object_set_data (GTK_OBJECT (item), "num_mails", num_mails);
	gtk_box_pack_end (GTK_BOX (hbox), num_mails, FALSE, FALSE, 0);
	gtk_widget_show (num_mails);

	unread_mails = mailbox_get_unread (mbox);
	sprintf (tmp, "%d", unread_mails);
	num_unread_mails = gtk_label_new (tmp);
	gtk_widget_set_name (num_unread_mails, "num_unread_mails");
	gtk_object_set_data (GTK_OBJECT (item), "num_unread_mails", num_unread_mails);
	gtk_box_pack_end (GTK_BOX (hbox), num_unread_mails, FALSE, FALSE, 10);
	gtk_widget_show (num_unread_mails);

	/* if there are any unread mails we underline the mailboxname.
	 * I would rather use bold, but how do you do that easily with a label? */
	if (unread_mails > 0) {
		i = 0;
		while (i < strlen (mboxname)) {
			tmp[i] = '_';
			i++;
		}
		tmp[i] = '\0';
		gtk_label_set_pattern (GTK_LABEL (label), tmp);
	}
	
	gtk_object_set_data (GTK_OBJECT (item), "mailboxformat", mbox);
	gtk_tree_append (GTK_TREE (GTK_TREE_ITEM_SUBTREE (tree)), item);
	gtk_widget_show (item);
	
	gtk_signal_connect (GTK_OBJECT (item), "button_press_event", 
			    GTK_SIGNAL_FUNC (on_treeBoxes_clicked), NULL);
	
	/* for unknown reason, this code is non-functional (drag-and-drop) */
	gtk_signal_connect (GTK_OBJECT (item), "drag_data_received",
			    GTK_SIGNAL_FUNC (on_treeBoxes_drag_data_recieved),
			    NULL);
	gtk_drag_dest_set (item, GTK_DEST_DEFAULT_ALL, target, 2, 
			   GDK_ACTION_MOVE | GDK_ACTION_COPY);
}

void refresh_mailboxes_fields (void)
{
	GtkWidget *treelocalBoxes = get_widget (frmMain, "treelocalBoxes");
	GtkWidget *mails, *unread_mails, *label;
	GList *items;
	Mailbox_t *mbox;
	int num_mails, num_unread_mails, i;
	char tmp[32];


	items = GTK_TREE (treelocalBoxes)->children;
	while (items != NULL) {
		mbox = gtk_object_get_data (GTK_OBJECT (items->data), "mailboxformat");
		num_mails = mailbox_get_size (mbox);
		num_unread_mails = mailbox_get_unread (mbox);
		mails = gtk_object_get_data (GTK_OBJECT (items->data), "num_mails");
		unread_mails = gtk_object_get_data (GTK_OBJECT (items->data), "num_unread_mails");
		label = gtk_object_get_data (GTK_OBJECT (items->data), "label");
		
		sprintf (tmp, "%d", num_mails);
		gtk_label_set_text (GTK_LABEL (mails), tmp);
		sprintf (tmp, "%d", num_unread_mails);
		gtk_label_set_text (GTK_LABEL (unread_mails), tmp);

		if (num_unread_mails > 0) {
			i = 0;
			while (i < strlen (mbox->name)) {
				tmp[i] = '_';
				i++;
			}
			tmp[i] = '\0';
			gtk_label_set_pattern (GTK_LABEL (label), tmp);
		} else {
			gtk_label_set_pattern (GTK_LABEL (label), "");
		}
		
		items = items->next;
	}
}

void refresh_mailboxes (int hilite)
{
	GtkWidget *treelocalBoxes = get_widget (frmMain, "treelocalBoxes_item");
	GtkWidget *treeremoteBoxes = get_widget (frmMain, "treeremoteBoxes_item");
	GtkWidget *treespoolBoxes = get_widget (frmMain, "treespoolBoxes_item");
	GtkWidget *item;
	GtkWidget *tree;
	GList *tmp_glist, *mailboxes_to_add = NULL;
	Mailbox_t *mbox;
	FILE *fp;
	DIR *mailbox_dir;
	struct dirent *mailbox_dirent;
	gchar *mailbox_name;
	gchar buffer[257];
	gint i, add_mailbox;
	
	if (refreshing_boxes)
		return;
	
	refreshing_boxes = TRUE;
	
	/* open spruce.conf so we can read mailbox info */
	if ((fp = fopen (spruceconf, "rt")) == NULL)
		return;
	
	NumFolders = 0;
	
	/* clear the mailbox trees */
	clear_mailbox_tree (treelocalBoxes, "treelocalBoxes");
	clear_mailbox_tree (treeremoteBoxes, "treeremoteBoxes");
	clear_mailbox_tree (treespoolBoxes, "treespoolBoxes");
	
	/* add default mailboxes to local mailbox tree */
	i = 0;
	while (*default_mailboxes[i] != '\0') {
		add_mailbox_to_tree (treelocalBoxes, MAILBOX_LOCAL, default_mailboxes[i]);
		NumFolders++;
		i++;
	}
	
	/* read remaining local mailbox tree from mailbox directory */
	
	/* read mailbox directory */
	if (!(mailbox_dir = opendir (mailboxdir))) {
		perror ("refresh_mailboxes(): opendir");
		return;
	}
	
	errno = 0;	/* Reset errno */
	while ((mailbox_dirent = readdir (mailbox_dir))) {
		/* Look for .mbx files */
		mailbox_name = g_strdup (mailbox_dirent->d_name);
		strrev (mailbox_name);
		if (!strncmp ("xbm.", mailbox_name, strlen ("xbm."))) {
			strrev (mailbox_name);
			
			/* Strip off the .mbx extension */
			*(mailbox_name + strlen (mailbox_name) - 4) = '\0';
			
			add_mailbox = TRUE;
			
			i = 0;
			/* Verify the mailbox we found is not a default */
			while (*default_mailboxes[i] != '\0') {
				if (!strcmp (default_mailboxes[i], mailbox_name)) {
					add_mailbox = FALSE;
					g_free (mailbox_name);
				}
				i++;
			}
			
			if (add_mailbox) {
				mailboxes_to_add =
				    g_list_insert_sorted (mailboxes_to_add,
							  mailbox_name,
							  (GCompareFunc) g_strcasecmp);
				
				NumFolders++;
			}
		} else {
			g_free (mailbox_name);
		}
		
		errno = 0;
	}
	
	if (errno) {
		perror ("refresh_mailboxes(): readdir");
		return;
	}
	
	closedir (mailbox_dir);
	
	g_list_foreach (mailboxes_to_add, (GFunc) add_local_mailbox_to_tree, NULL);
	
	tmp_glist = mailboxes_to_add;
	while (tmp_glist) {
		g_free (tmp_glist->data);
		tmp_glist = tmp_glist->next;
	}
	g_list_free (mailboxes_to_add);
	
	/* read mailbox info (spool and remote) from spruce.conf */
	while (!feof (fp)) {
		memset (buffer, 0, sizeof (buffer));
		fgets (buffer, sizeof (buffer)-1, fp);
		g_strchug (buffer);
		
		if (!strncmp ("mbox-spool", buffer, strlen ("mbox-spool"))) {
			strcut (buffer, 0, strlen ("mbox-spool"));
			g_strstrip (buffer);
			add_mailbox_to_tree (treespoolBoxes, MAILBOX_SPOOL, buffer);
			continue;
		}
	}
	
	/* setup the remote mailbox trees */
	for (i = 0; i < g_list_length (ImapAccounts); i++) {
		mbox = g_malloc0 (sizeof (Mailbox_t));
		mbox->type = MAILBOX_IMAP;
		mbox->imap = g_list_nth_data (ImapAccounts, i);
		item = gtk_tree_item_new_with_label (mbox->imap->server.hostname);
		gtk_object_set_data_full (GTK_OBJECT (item), "mbox", mbox, g_free);
		gtk_widget_show (item);
		gtk_tree_append (GTK_TREE (GTK_TREE_ITEM_SUBTREE (treeremoteBoxes)), item);
		
		tree = gtk_tree_new ();
		gtk_tree_item_set_subtree (GTK_TREE_ITEM (item), tree);
		gtk_widget_show (tree);
		gtk_signal_connect (GTK_OBJECT (item), "expand",
				    GTK_SIGNAL_FUNC (on_imaptree_expand),
				    (gpointer) mbox);
		gtk_signal_connect (GTK_OBJECT (item), "collapse",
				    GTK_SIGNAL_FUNC (on_imaptree_collapse),
				    (gpointer) mbox);
		gtk_signal_connect (GTK_OBJECT (item), "button_press_event",
				    GTK_SIGNAL_FUNC (on_treeBoxes_clicked),
				    NULL);
	}
	
	fclose (fp);
	
	/* if there are any folders... */
	if (NumFolders) {
		if (hilite == TOP) {
			gtk_tree_select_item (GTK_TREE (GTK_TREE_ITEM (treelocalBoxes)->subtree), 0);
			cMbox = 0;
		} else if (hilite == STAY) {
			if (cMbox >= NumFolders)
				cMbox = NumFolders - 1;
		}
		/*refresh_mail_list (TOP); */
	} else
		cMbox = -1;

	refreshing_boxes = FALSE;
}

static void refresh_mail_list_elem (gpointer data, gpointer user_data)
{
	static gchar *text[] = {"Id", "Sender", "Subject", "Date"};
	Index_t *index = (Index_t *) data;
	guint32 *i = (guint32 *) user_data;
	GtkWidget *clist = get_widget (frmMain, "clistMail");
	gchar id[10], *sender, *subject, *date;
	GtkStyle *style1, *style2;
	
	g_return_if_fail (index);
	g_return_if_fail (i);
	
	g_snprintf (id, 10, "%d", *i);

	sender = index_get_sender (index);
	subject = index_get_subject (index);
	date = index_get_date (index);
	
	gtk_clist_append (GTK_CLIST (clist), (gchar **) text);
	gtk_clist_set_text (GTK_CLIST (clist), *i - 1, 0, id);
	gtk_clist_set_text (GTK_CLIST (clist), *i - 1, 1, sender ? sender : "");
	gtk_clist_set_text (GTK_CLIST (clist), *i - 1, 2, subject ? subject : "");
	gtk_clist_set_text (GTK_CLIST (clist), *i - 1, 3, date ? date : "");
	
	style1 = gtk_widget_get_style (clist);
	style2 = gtk_style_copy (style1);
	
	if (!unread_mesg_header_font || !read_mesg_header_font) {
		(*i)++;
		return;
	}
	
	style2->font = gdk_font_load (index_get_seen (index) ? read_mesg_header_font : unread_mesg_header_font);
	if (style2->font)
		gtk_clist_set_row_style (GTK_CLIST (clist), *i - 1, style2);
	
	gtk_style_unref (style2);
	(*i)++;
}

void refresh_mail_list (gint hilite)
{
	GtkWidget *clist = get_widget (frmMain, "clistMail");
	guint32 i = 1, max = 0;
	Mailbox_t *mbox;
	GList *index_list;
	
	if (!selected_mailbox) {
		gtk_clist_clear (GTK_CLIST (clist));
		return;
	}
	
	g_return_if_fail (!refreshing_maillist);
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	
	refreshing_maillist = TRUE;
	gtk_clist_freeze (GTK_CLIST (clist));
	gtk_clist_clear (GTK_CLIST (clist));
	
	max = mailbox_get_size (mbox);
	index_list = mailbox_get_index_list (mbox);
	g_list_foreach (index_list, refresh_mail_list_elem, &i);

	/* if there are any messages... */
	if (max > 0) {
		if (hilite == TOP)
			cMesg = 0;
		else
			if (hilite == STAY) {
				if (cMesg >= max)
					cMesg = max - 1;
			}
		gtk_clist_sort (GTK_CLIST (clist));  /* sorts the clist to the default column */
		/*refresh_txtMessage ();*/
		cMesg--;
		gtk_clist_select_row (GTK_CLIST (clist), cMesg + 1, 0);
		gtk_clist_moveto (GTK_CLIST (clist), cMesg, 0, 1.0, 0.0);
	} else {
		cMesg = -1;
		refresh_txtMessage ();
	}
	
	NumMesgs = max;
	
	/*gtk_clist_sort (GTK_CLIST (clist));*/  /* sorts the clist to the default column */
	gtk_clist_thaw (GTK_CLIST (clist));
	refreshing_maillist = FALSE;
}

void refresh_txtMessage ()
{
	GtkWidget *txtBody      = get_widget (frmMain, "txtMessage");
	GtkWidget *clistMail    = get_widget (frmMain, "clistMail");
	GtkWidget *clistMime    = get_widget (frmMain, "clistMime");
	gint part_len;
	guint len;
	gchar *header_mesg = NULL;
	gchar *msg = NULL;
	gchar *id;
	gchar *part = NULL;
	GdkFont *fixed_font = NULL;
	Mailbox_t *mbox;
	Index_t *index;
	
	mime = 0;   
	
	g_return_if_fail (!refreshing_message);
	refreshing_message = TRUE;
	gtk_text_freeze (GTK_TEXT (txtBody));

	/* delete all the the text in there */
	len = gtk_text_get_length (GTK_TEXT (txtBody));
	gtk_text_set_point (GTK_TEXT (txtBody), len);
	gtk_text_backward_delete (GTK_TEXT (txtBody), len);
	
	/* and clear the mime clist */
	if (mime_preferred_part == 0)
		gtk_clist_clear (GTK_CLIST (clistMime));
	
	if (cMesg >= 0) {
		mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
		g_return_if_fail (mbox);
		
		/* put the new email in there */
		gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);
		
		index = mailbox_get_index(mbox, atoi (id));
		msg = mailbox_get_msg (mbox, index, &mime, 2);
		header_mesg = mailbox_get_msg (mbox, index, NULL, 
					       hdr_trim_level);
		
		/* Load a fixed font */
		if (mesg_body_font)
			fixed_font = gdk_font_load (mesg_body_font);
		
		if ((hdr_trim_level != HEADER_TRIMLEVEL_KEEP_NONE) && 
		    (mime_preferred_part == 0)) {
			GdkColormap *cmap;
			GdkColor colour;
			gchar *hdrstripped;
			guint hdrlen;
			
			/* Get the system colour map and allocate the colour blue */
			cmap = gdk_colormap_get_system();
			colour.red = 0;
			colour.green = 0;
			colour.blue = 0xffff;
			if (!gdk_color_alloc(cmap, &colour))
				g_error ("couldn't allocate colour");
			
			hdrlen = mesg_get_header_len (header_mesg);
			hdrstripped = mesg_get_trimmed_header (header_mesg, hdr_trim_level, hdrlen);
			gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL, hdrstripped, -1);
			gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL, "\n\n", -1);
			g_free (hdrstripped);
		}
		
		if (mime) {
			if (mime_preferred_part == 0)
				part = mailbox_get_msg_mime_part (mbox, index, mime_get_first_text (), &part_len);
			else
				part = mailbox_get_msg_mime_part (mbox, index, mime_preferred_part, &part_len);
			
			if (part) {
#ifdef PGP_PROGRAM
				gchar *decrypted = NULL;
				GdkColormap *cmap;
				GdkColor colour;
				
				if (pgp_detect (part)) {
					SpruceException *ex;
					
					ex = spruce_exception_new ();
					decrypted = pgp_decrypt (part, pgp_passphrase, ex);
					
					if (decrypted) {
						/* failed to decrypt - highlight in red */
						/* Get the system colour map and allocate the colour red */
						cmap = gdk_colormap_get_system();
						colour.red = 0xffff;
						colour.green = 0;
						colour.blue = 0;
						
						if (!gdk_color_alloc(cmap, &colour))
							g_error ("couldn't allocate colour");
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL,
						     "----- Begin Decrypted PGP Message -----\n\n", -1);
						
						gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, decrypted, -1);
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL,
						     "\n----- End Decrypted PGP Message -----\n", -1);
						
						g_free (part);
						g_free (decrypted);
						spruce_exception_free (ex);
					} else {
						/* failed to decrypt - highlight in red */
						/* Get the system colour map and allocate the colour red */
						GtkWidget *dialog;
						
						cmap = gdk_colormap_get_system ();
						colour.red = 0xffff;
						colour.green = 0;
						colour.blue = 0;
						
						if (!gdk_color_alloc (cmap, &colour))
							g_error ("couldn't allocate colour");
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL, part, -1);
						
						g_free (part);
						
						dialog = spruce_dialog_new_from_exception (ex);
						gtk_widget_show (dialog);
						spruce_exception_free (ex);
					}
				} else {
					gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, part, -1);
					g_free (part);
				}
#else
				gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, part, -1);
				g_free (part);
#endif /* PGP_PROGRAM */
			} else {
				/* no mime parts but yes to mime */
#ifdef PGP_PROGRAM
				gchar *decrypted = NULL;
				GdkColormap *cmap;
				GdkColor colour;
				
				/*trim_header (msg, 0);*/
				if (pgp_detect (msg)) {
					SpruceException *ex;
					
					ex = spruce_exception_new ();
					decrypted = pgp_decrypt (msg, pgp_passphrase, ex);
					
					if (decrypted) {
						/* failed to decrypt - highlight in red */
						/* Get the system colour map and allocate the colour red */
						cmap = gdk_colormap_get_system ();
						colour.red = 0xffff;
						colour.green = 0;
						colour.blue = 0;
						
						if (!gdk_color_alloc (cmap, &colour))
							g_error ("couldn't allocate colour");
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL,
						     "----- Begin Decrypted PGP Message -----\n\n", -1);
						
						gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, decrypted, -1);
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL,
						     "\n----- End Decrypted PGP Message -----\n", -1);
						
						g_free (decrypted);
						spruce_exception_free (ex);
					} else {
						/* failed to decrypt - highlight in red */
						/* Get the system colour map and allocate the colour red */
						GtkWidget *dialog;
						
						cmap = gdk_colormap_get_system ();
						colour.red = 0xffff;
						colour.green = 0;
						colour.blue = 0;
						
						if (!gdk_color_alloc (cmap, &colour))
							g_error ("couldn't allocate colour");
						
						gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL, msg, -1);
						
						dialog = spruce_dialog_new_from_exception (ex);
						gtk_widget_show (dialog);
						spruce_exception_free (ex);
					}
				} else {
					gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, msg, -1);
				}
#else
				g_free (msg);
				msg = mailbox_get_msg (mbox, index, NULL, 0);
				gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, msg, -1);
#endif /* PGP_PROGRAM */
			}
		} else {
			/* no mime at all */
#ifdef PGP_PROGRAM
			gchar *decrypted = NULL;
			GdkColormap *cmap;
			GdkColor colour;
			
			/*trim_header (msg, 0);*/
			g_free (msg);
			msg = mailbox_get_msg (mbox, index, NULL, 0);
			if (pgp_detect (msg)) {
				SpruceException *ex;
				
				ex = spruce_exception_new ();
				decrypted = pgp_decrypt (msg, pgp_passphrase, ex);
				
				if (decrypted) {
					gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, decrypted, -1);
					g_free (decrypted);
					spruce_exception_free (ex);
				} else {
					/* failed to decrypt - highlight in red */
					/* Get the system colour map and allocate the colour red */
					GtkWidget *dialog;
					
					cmap = gdk_colormap_get_system ();
					colour.red = 0xffff;
					colour.green = 0;
					colour.blue = 0;
					
					if (!gdk_color_alloc (cmap, &colour))
						g_error ("couldn't allocate colour");
					
					gti (GTK_TEXT (txtBody), fixed_font, &colour, NULL, msg, -1);
					
					dialog = spruce_dialog_new_from_exception (ex);
					gtk_widget_show (dialog);
					spruce_exception_free (ex);
				}
			} else {
				gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, msg, -1);
			}
#else
			g_free (msg);
			msg = mailbox_get_msg (mbox, index, NULL, 0);
			gti (GTK_TEXT (txtBody), fixed_font, NULL, NULL, msg, -1);
#endif /* PGP_PROGRAM */
		}
		
		if (mime_preferred_part == 0 && mime)
			refresh_clistMime ();
		
		g_free (msg);
		g_free (header_mesg);
	}
	
	gtk_text_thaw (GTK_TEXT (txtBody));
	mime_preferred_part = 0;
	refreshing_message = FALSE;
}

void refresh_clistMime ()
{
	GtkWidget *clist = get_widget (frmMain, "clistMime");
	GList *tmp;
	gchar *text[4], id[16], len[32], type[128], name[128];
	gint i, ret;
	
	gtk_clist_clear (GTK_CLIST (clist));
   
	if (g_list_length (mime_parts) > 1) {
		i = 1;
		tmp = mime_parts;
		while (tmp != NULL) {
			sprintf (id, "%d", i);
			text[0] = id;
			sprintf (type, "%s/%s", ((struct mime_part *)tmp->data)->type,
				 ((struct mime_part *)tmp->data)->subtype);
			text[1] = type;
			ret = mime_get_parameter_value (((struct mime_part *)tmp->data)->parameter, "name", name);
			if (ret)
				text[2] = name;
			else
				text[2] = NULL;
			sprintf (len, "%d", ((struct mime_part *)tmp->data)->len);
			text[3] = len;
			gtk_clist_append (GTK_CLIST (clist), text);
			i++;
			tmp = tmp->next;
		}
	}
}

void clistMail_update ()
{
	refresh_mail_list (TRUE);
}

void read_sizes (void)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GtkWidget *pan1 = get_widget (frmMain, "pan1");
	GtkWidget *pan2 = get_widget (frmMain, "pan2");
	GtkWidget *pan3 = get_widget (frmMain, "pan3");
	FILE *fp;
	gchar buffer[256], *field;
	gint width, heigth, i, size;
	
	if (save_sizes) {
		fp = fopen (spruceconf, "rt");
		if (fp == NULL) {
			fprintf (stderr, "Spruce: %s failed to open for reading.\n", spruceconf);
			fprintf (stderr, "Please check with your local Administrator.\n");
			return;
		}
		
		while (!feof (fp)) {
			fgets (buffer, 256, fp);
			g_strchug (buffer);
			if (buffer[0] != '#' && buffer[0] != '\n') {
				if (!strncmp ("spruce-sizes", buffer, strlen ("spruce-sizes"))) {
					strcut (buffer, 0, strlen ("spruce-sizes"));
					g_strstrip (buffer);
					if (!strcmp ("",buffer))
						continue;
					
					field = get_field (buffer, 1, ':');
					width = atoi (field);
					g_free (field);
					
					field = get_field (buffer, 2, ':');
					heigth = atoi (field);
					g_free (field);
					
					gtk_widget_set_usize (frmMain, width, heigth);
					
					i = 0;
					/* read column sizes... */
					while (i < GTK_CLIST (clistMail)->columns) {
						field = get_field (buffer, i+3, ':');
						gtk_clist_set_column_width (GTK_CLIST (clistMail), i, atoi (field));
						g_free (field);
						i++;
					}

					/* read pane sizes... */
					field = get_field (buffer, i+3, ':');
					size = atoi (field);
					if (size)
						gtk_paned_set_position (GTK_PANED (pan1), size);
					g_free (field);
					i++;
					
					field = get_field (buffer, i+3, ':');
					size = atoi (field);
					if (size)
						gtk_paned_set_position (GTK_PANED (pan2), size);
					g_free (field);
					i++;
					
					field = get_field (buffer, i+3, ':');
					size = atoi (field);
					if (size)
						gtk_paned_set_position (GTK_PANED (pan3), size);
					g_free (field);
					i++;

					/* read compose window size... */
					field = get_field (buffer, i+3, ':');
					size = atoi (field);
					compose_window_width = size ? size : DEFAULT_COMPOSE_WINDOW_WIDTH;
					g_free (field);
					i++;

					field = get_field (buffer, i+3, ':');
					size = atoi (field);
					compose_window_height = size ? size : DEFAULT_COMPOSE_WINDOW_HEIGHT;
					g_free (field);
				}
			}
		}   
	}
}

void read_column_sort (void)
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	FILE *fp;
	gchar buffer[257], *field;
	
	/* default values */
	sort_column = 0;
	sort_type = GTK_SORT_ASCENDING;
	
	/* get "real" values */
	if (TRUE) {
		fp = fopen (spruceconf, "rt");
		if (fp == NULL) {
			fprintf (stderr, "Spruce: %s failed to open for reading.\n", spruceconf);
			fprintf (stderr, "Please check with your local Administrator.\n");
			return;
		}
		
		while (!feof (fp)) {
			memset (buffer, 0, sizeof (buffer));
			fgets (buffer, sizeof (buffer)-1, fp);
			g_strchug (buffer);
			if (buffer[0] != '#' && buffer[0] != '\n') {
				if (!strncmp ("column-sort", buffer, strlen ("column-sort"))) {
					strcut (buffer, 0, strlen ("column-sort"));
					g_strstrip (buffer);
					if (!*buffer)
						continue;
					
					field = get_field (buffer, 1, ':');
					sort_column = atoi (field);
					g_free (field);
					
					field = get_field (buffer, 2, ':');
					sort_type = atoi (field);
					g_free (field);
					
					/* sort the list now */
					gtk_clist_freeze (GTK_CLIST (clistMail));
					gtk_clist_set_sort_column (GTK_CLIST (clistMail), sort_column);
					gtk_clist_set_sort_type (GTK_CLIST (clistMail), sort_type);
					switch (sort_column) {
					case 0:         /* sort by message id */
						gtk_clist_set_compare_func (GTK_CLIST (clistMail), id_compare);
						break;
					case 3:         /* sort by date */
						gtk_clist_set_compare_func (GTK_CLIST (clistMail), date_compare);
						break;
					default:
						gtk_clist_set_compare_func (GTK_CLIST (clistMail), str_compare);
					}
					
					gtk_clist_sort (GTK_CLIST (clistMail));
					gtk_clist_thaw (GTK_CLIST (clistMail));
				}
			}
		}   
	}
}

void update_messageStatus ()
{
	GtkWidget *clistMail = get_widget (frmMain, "clistMail");
	GtkStyle *rowstyle;
	GtkStyle *newrowstyle;
	Mailbox_t *mbox;
	Index_t *index;
	gchar *id;
	
	g_return_if_fail (selected_mailbox);
	
	mbox = gtk_object_get_data (GTK_OBJECT (selected_mailbox), "mailboxformat");
	g_return_if_fail (mbox);
	g_return_if_fail (cMesg >= 0);
	
	gtk_clist_get_text (GTK_CLIST (clistMail), cMesg, 0, &id);   
	
	index = mailbox_get_index (mbox, atoi (id));
	
	if (index_get_seen (index))
		return;
	
	index_set_seen (index, TRUE);
	g_return_if_fail (read_mesg_header_font);
	
	rowstyle = gtk_clist_get_row_style (GTK_CLIST (clistMail), cMesg);
	g_return_if_fail (rowstyle);
	
	newrowstyle = gtk_style_copy (rowstyle);
	g_return_if_fail (newrowstyle);
	
	newrowstyle->font = gdk_font_load (read_mesg_header_font);
	if (newrowstyle->font)
		gtk_clist_set_row_style (GTK_CLIST (clistMail), cMesg, newrowstyle);
	gtk_style_unref (newrowstyle);

	/* since this was an unread mail, we also update the mail counts in the mailbox pane */
	refresh_mailboxes_fields ();
}
