/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include "mmap.h"

void *g_mmap (void *start, size_t length, gint prot, gint flags, gint fd, off_t offset)
{
#if defined(HAVE_MMAP) && defined(HAVE_MUNMAP)
	return mmap (start, length, prot, flags, fd, offset);
#else
	void *buffer;
	size_t size;
	struct stat st;
	
	if (fstat (fd, &st) == -1)
		return MAP_FAILED;
	
	lseek (fd, offset, SEEK_SET);
	
	buffer = g_malloc (length);
	
	size = (st.st_size - offset) < length ? st.st_size - offset : length;
	
	read (fd, buffer, size);
	
	return buffer;
#endif
}

gint g_munmap (void *start, size_t length)
{
#if defined(HAVE_MUNMAP) && defined(HAVE_MMAP)
	return munmap (start, length);
#else
	if (start != MAP_FAILED)
		g_free (start);
	start = NULL;

	return 0;
#endif
}
