/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>

#include "mmap.h"
#include "move.h"

gint move (const gchar *oldpath, const gchar *newpath)
{
	gchar *ptr, *contents;
	struct stat st;
	gint oldfd, newfd;
	guint32 len;
	
	oldfd = open (oldpath, O_RDONLY);
	if (oldfd == -1)
		return -1;
	
	if (fstat (oldfd, &st) == -1) {
		close (oldfd);
		return -1;
	}
	
	if (st.st_size) {
		ptr = (gchar *) g_mmap (0, st.st_size, PROT_READ, MAP_PRIVATE, oldfd, 0);
		if (ptr == (gchar *)MAP_FAILED) {
			close (oldfd);
			return -1;
		}
		
		newfd = open (newpath, O_WRONLY | O_CREAT | O_TRUNC, 0600);
		if (newfd == -1) {
			close (oldfd);
			g_munmap (ptr, st.st_size);
			return -1;
		}
		
		contents = g_strndup (ptr, st.st_size);
		len = strlen (contents);
		
		if (write (newfd, contents, len) != len) {
			g_free (contents);
			close (newfd);
			g_munmap (ptr, st.st_size);
			close (oldfd);
			/*remove (newpath);*/
			return -1;
		}
		g_free (contents);
		close (newfd);
		g_munmap (ptr, st.st_size);
	} else {
		newfd = open (newpath, O_WRONLY | O_CREAT | O_TRUNC, 0600);
		if (newfd == -1) {
			close (oldfd);
			return -1;
		}
		close (newfd);
	}
	
	close (oldfd);
	
	remove (oldpath);
	
	return 0;
}
