/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef __PARSE_H__
#define __PARSE_H__

#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus */

#include <glib.h>
#undef MIN
#undef MAX
#include <stdio.h>
#include <string.h>
#include "mime.h"

/* getopts_long argument type */
typedef enum { no_arg, required_arg, optional_arg } golarg_t;

typedef struct {
	const gchar *name;
	golarg_t has_arg;
	/*gint *flag;*/
	gint val;
} longopts_t;

GList *g_list_strtok (gchar *string, gchar *delimeters);

/* returns the position of the string or -1 if not found */
gint find_string (const gchar *haystack, const gchar *needle);

/* An implementation of Boyer Moore's search string algorythm */
#define _nstrstr(a, b) nstrstr((a), (b), FALSE)
gchar *nstrstr (gchar *haystack, gchar *needle, gboolean recyle);

/* temporary replacement for strcasestr() since it isn't a standard function apparently */
gchar *strstrcase (gchar *haystack, gchar *needle);

/* a bound strstr */
gchar *strstrbound (gchar *haystack, gchar *needle, gchar *end);

/* reverses the order of the chars in the string */
gchar *strrev (gchar *string);

/* inserts a string in the "pos" position... returns 1 on success or 0 on fail */
#define insert_string(a, b, c)   strinsert((a), (b), (c))
gint strinsert (gchar *string, const gchar *insertstr, guint pos);

/* inserts a string in the "pos" position... returns 1 on success or 0 on fail */
gint strninsert (gchar *string, const gchar *insertstr, guint pos, size_t strmax);

/* removes the numchars number of chars starting at index... returns 1 on succes or 0 on fail */
#define remove_string(a, b, c)   strcut((a), (b), (c))
gint strcut (gchar *string, guint index, guint numchars);

/* replaces 1 string with another string (first occurence only) */
#define string_replace(a, b, c) strrep((a), (b), (c))
gboolean strrep (gchar *string, const gchar *remove, const gchar *replace);

/* trims all the spaces and tabs from the leading edge of the string */
gchar *trim_leadspc (gchar *string);

/* trims all the trailing spaces/tabs/newlines from the string */
gchar *trim_tailspc (gchar *string);

/* trims all the spaces/tabs/newlines from lead/tail of the string */
gchar *trim_whtspc (gchar *string);

/* unquote a quoted string */
gchar *unquote (gchar *string);

/* strips every occurance of a char in the string */
gchar *strip (gchar *string, gchar c);

/* change all commas (,) between addresses to semi-colons (;) */
gchar *parse_addr_list (gchar *list);

/* change all semi-colons (;) between addresses to commas (,) */
gchar *unparse_addr_list (gchar *list);

/* returns the number of fields in a string separated by the delimeter */
guint get_num_fields (gchar *string, gchar delim);

/* returns the field in the string given the delimiter */
gchar *get_field (gchar *string, guint field, gchar delim);

/* decode the 8-bit text */
gint parse_8bit (gchar *text);

/* 8-bit encode text */
gchar *make_8bit (guchar *text);

/* Return the length of the longest line in a message */
guint mesg_get_longest_line_len (gchar *mesg, guint tab_equivalent);

/* Return the count of tabs found in a message */
guint mesg_get_tab_count (gchar *mesg);

/* wrap without breaking up words */
gchar *soft_word_wrap (gchar *text, guint len, guint tab_equivalent);

/* word wrap a body of text to `len` chars per line */
gchar *word_wrap (gchar *text, guint len, guint tab_equivalent, 
		  guint tab_convert, guint det_pgp_clearsign);

/* quote a body of text by placing a "> " before each line */
gchar *quote_body (gchar *body, gchar *rcpt, gchar *date);

/* parse command-line for option chars */
gint getopts_short (gint argc, gchar *argv[], gchar *optstring);

/* parse the command-line for long options */
gint getopts_long (gint argc, gchar *argv[], gchar *optstring,
                   longopts_t *longopts, gint *longindex, gchar **param);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
