/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* spruce-execpetion.h : exception utils */
/* 
 *
 * Authors: Bertrand Guiheneuf <bertrand@helixcode.com>
 *          Jeffrey Stedfast <fejj@helixcode.com>
 *
 * Copyright 1999, 2000 Helix Code, Inc. (http://www.helixcode.com)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */


#ifndef __SPRUCE_EXCEPTION_H__
#define __SPRUCE_EXCEPTION_H__


#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus */

#include <glib.h>

typedef enum {
#include "spruce-exception-list.def"

} ExceptionId;


struct _SpruceException {
	/* do not access the fields directly */
	ExceptionId id;
	char *desc;
};

typedef struct _SpruceException SpruceException;

/* creation and destruction functions */
SpruceException *         spruce_exception_new           (void);
void                      spruce_exception_free          (SpruceException *exception);
void                      spruce_exception_init          (SpruceException *ex);


/* exception content manipulation */
void                      spruce_exception_clear         (SpruceException *exception);
void                      spruce_exception_set           (SpruceException *ex,
							 ExceptionId id,
							 const char *desc);
void                      spruce_exception_setv          (SpruceException *ex,
							 ExceptionId id,
							 const char *format,  
							 ...);


/* exception content transfer */
void                      spruce_exception_xfer          (SpruceException *ex_dst,
							 SpruceException *ex_src);


/* exception content retrieval */
ExceptionId               spruce_exception_get_id        (SpruceException *ex);
const gchar *             spruce_exception_get_description (SpruceException *ex);

#define spruce_exception_is_set(ex) (spruce_exception_get_id (ex) != SPRUCE_EXCEPTION_NONE)

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __SPRUCE_EXCEPTION_H__ */
