/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  Authors: George Talusan
 *           Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "uid.h"

static GHashTable *uid_table = NULL;
extern gchar *uid_file;

void uid_init (void)
{
	gchar line[512];
	
	FILE *fp;
	
	if (uid_table) {
		uid_free ();
		uid_table = NULL;
	}
	
	uid_table = g_hash_table_new (g_str_hash, g_str_equal);
	
	fp = fopen (uid_file, "rt");
	
	if (!fp)
		return;
	
	while (!feof (fp)) {
		memset (line, 0, sizeof (line));
		fgets (line, sizeof (line)-1, fp);
		g_strstrip (line);
		if (*line)
			g_hash_table_insert (uid_table, g_strdup (line), GINT_TO_POINTER (1));
	}
	fclose (fp);
}

static gboolean uid_elem_free (gpointer key, gpointer value, gpointer user_data)
{
	FILE *fp = (FILE *) user_data;
	
	fprintf (fp, "%s\n", (gchar *) key);
	g_free (key);
	
	return TRUE;
}

void uid_free (void)
{
	FILE *fp;
	
	fp = fopen (uid_file, "wt");
	g_return_if_fail (fp);
	
	g_hash_table_foreach_remove (uid_table, uid_elem_free, fp);
	g_hash_table_destroy (uid_table);
	uid_table = NULL;
	
	fclose (fp);
}

void uid_add (gchar *uid)
{
	g_return_if_fail (uid_table);
	
	g_strstrip (uid);
	g_hash_table_insert (uid_table, g_strdup (uid), GINT_TO_POINTER (1));
}

gboolean uid_lookup(gchar *uid)
{
	g_return_val_if_fail (uid_table, FALSE);
	
	g_strstrip (uid);
	
	return (gboolean) g_hash_table_lookup (uid_table, uid);
}
