/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>

static char *xml = "<?xml version=\"1.0\"?>\n<efolder>\n   <type>mail</type>\n</efolder>\n";

char *opath = NULL;

static char *
strndup (const char *blob, int len)
{
	char *new;
	
	new = calloc (sizeof (char), len + 1);
	strncpy (new, blob, len);
	
	return new;
}

static void
spruce_to_mbox (const char *spruce)
{
	char *mailbox;
	char *mbox;
	char *dirname;
	char *metadata;
	char *p, *q, *r;
	struct stat st;
	int sprucefd, mboxfd, metafd;
	
	/* create the base mailbox name (no .mbx extension) */
	mailbox = strndup (spruce, strlen (spruce) - 4);
	
	if (stat (spruce, &st) == -1 || !st.st_size) {
		perror ("spruce mailbox::stat ()");
		free (mailbox);
		
		return;
	}
	
	if ((sprucefd = open (spruce, O_RDONLY)) == -1) {
		perror ("spruce mailbox::open ()");
		free (mailbox);
		
		return;
	}
	
	/* make the directory to place the mbox and metadata files */
	dirname = calloc (sizeof (char), strlen (opath) + 1 + strlen (mailbox) + 1);
	sprintf (dirname, "%s/%s", opath, mailbox);
	free (mailbox);
	
	if (mkdir (dirname, 0700) == -1 && errno != EEXIST) {
		perror ("mkdir ()");
		free (dirname);
		close (sprucefd);
		
		return;
	}
	
	/* write out the folder-metadata (used by Evolution) */
	metadata = calloc (sizeof (char), strlen (dirname) + 20);
	sprintf (metadata, "%s/folder-metadata.xml", dirname);
	if ((metafd = open (metadata, O_CREAT | O_TRUNC | O_WRONLY, 0600)) == -1) {
		perror ("folder-metadata.xml::open ()");
		free (metadata);
		close (sprucefd);
		
		return;
	}
	free (metadata);
	write (metafd, xml, strlen (xml));
	close (metafd);
	
	mbox = calloc (sizeof (char), strlen (dirname) + 6);
	sprintf (mbox, "%s/mbox", dirname);
	free (dirname);
	
	if ((mboxfd = open (mbox, O_CREAT | O_TRUNC | O_WRONLY, 0600)) == -1) {
		perror ("mbox::open ()");
		free (mbox);
		close (sprucefd);
		
		return;
	}
	free (mbox);
	
	p = q = (char *) mmap (0, st.st_size, PROT_READ, MAP_PRIVATE, sprucefd, 0);
	
	if (p == (char *) MAP_FAILED) {
		perror ("spruce mailbox::mmap ()");
		close (sprucefd);
		close (mboxfd);
		
		return;
	}
	
	/* p and r point to the beginning and end of an email message */
	while ((r = strstr (p, "\n.\r\n"))) {
		char *raw;
		
		raw = strndup (p, (int) (r + 1 - p));
		
		if (write (mboxfd, "From -\n", 7) != 7) {
			perror ("write ()");
			fprintf (stderr, "Could not finish converting '%s' to mbox format.\n", spruce);
			free (raw);
			munmap (q, st.st_size);
			close (sprucefd);
			close (mboxfd);
			
			return;
		}
		
		if (write (mboxfd, raw, strlen (raw)) != strlen (raw)) {
			perror ("write ()");
			fprintf (stderr, "Could not finish converting '%s' to mbox format.\n", spruce);
			free (raw);
			munmap (q, st.st_size);
			close (sprucefd);
			close (mboxfd);
			
			return;
		}
		
		free (raw);
		
		p = r + 4; /* skip over the \n.\r\n */
	}
	
	munmap (q, st.st_size);
	close (sprucefd);
	close (mboxfd);
}

int main (int argc, char **argv)
{
	int i;
	
	if (argc < 3) {
		printf ("Usage: %s output-path *.mbx\n", argv[0]);
		return 0;
	}
	
	opath = argv[1];
	
	for (i = 2; i < argc; i++)
		spruce_to_mbox (argv[i]);
	
	return 0;
}
